from ecoscope.io import earthranger, eetools, raster, utils
from ecoscope.io.earthranger import EarthRangerIO
from ecoscope.io.utils import download_file

__all__ = [
    "earthranger",
    "EarthRangerIO",
    "download_file",
    "eetools",
    "raster",
    "utils",
]



import ecoscope


def test_speed_geoseries(movbank_relocations):
    trajectory = ecoscope.base.Trajectory.from_relocations(movbank_relocations)
    sdf = ecoscope.analysis.speed.SpeedDataFrame.from_trajectory(trajectory)
    assert not sdf.geometry.is_empty.any()
    assert not sdf.geometry.isna().any()



import os
from tempfile import NamedTemporaryFile

import geopandas as gpd
import geopandas.testing
import numpy as np
import pytest

import ecoscope


@pytest.mark.skip(reason="this has been failing since May 2022; will be fixed in a follow-up pull")
def test_etd_range(movbank_relocations):
    # apply relocation coordinate filter to movbank data
    pnts_filter = ecoscope.base.RelocsCoordinateFilter(
        min_x=-5,
        max_x=1,
        min_y=12,
        max_y=18,
        filter_point_coords=[[180, 90], [0, 0]],
    )
    movbank_relocations.apply_reloc_filter(pnts_filter, inplace=True)
    movbank_relocations.remove_filtered(inplace=True)

    # Create Trajectory
    movbank_trajectory_gdf = ecoscope.base.Trajectory.from_relocations(movbank_relocations)

    raster_profile = ecoscope.io.raster.RasterProfile(
        pixel_size=250.0,
        crs="ESRI:102022",
        nodata_value=np.nan,
        band_count=1,  # Albers Africa Equal Area Conic
    )

    file = NamedTemporaryFile(delete=False)
    try:
        ecoscope.analysis.UD.calculate_etd_range(
            trajectory_gdf=movbank_trajectory_gdf,
            output_path=file.name,
            max_speed_kmhr=1.05 * movbank_trajectory_gdf.speed_kmhr.max(),
            raster_profile=raster_profile,
            expansion_factor=1.3,
        )

        percentile_area = ecoscope.analysis.get_percentile_area(
            percentile_levels=[99.9], raster_path=file.name, subject_id="Salif_Keita"
        ).to_crs(4326)
    finally:
        file.close()
        os.unlink(file.name)

    expected_percentile_area = gpd.read_feather("tests/test_output/etd_percentile_area.feather")
    gpd.testing.assert_geodataframe_equal(percentile_area, expected_percentile_area, check_less_precise=True)


@pytest.mark.skip(reason="this has been failing since May 2022; will be fixed in a follow-up pull")
def test_reduce_regions(aoi_gdf):
    raster_names = ["tests/sample_data/raster/mara_dem.tif"]
    result = ecoscope.io.raster.reduce_region(aoi_gdf, raster_names, np.mean)
    assert result[raster_names[0]].sum() > 0



import ee
import pandas as pd
import pytest

import ecoscope

if not pytest.earthengine:
    pytest.skip("Skipping tests because connection to Earth Engine is not available.", allow_module_level=True)


def test_albedo_anomaly(aoi_gdf):
    tmp_gdf = aoi_gdf.to_crs(4326)

    params = {
        "img_coll": ee.ImageCollection("MODIS/006/MCD43C3").select("Albedo_BSA_vis"),
        "historical_start": "2000-01-01",
        "start": "2010-01-01",
        "end": "2022-01-01",
        "scale": 5000.0,
    }

    result = ecoscope.io.eetools.chunk_gdf(
        gdf=tmp_gdf,
        label_func=ecoscope.io.eetools.calculate_anomaly,
        label_func_kwargs=params,
        df_chunk_size=1,
        max_workers=5,
    )

    assert isinstance(result, pd.DataFrame)
    assert result["Albedo_BSA_vis"].mean() > 0


def test_label_gdf_with_temporal_image_collection_by_features_aois(aoi_gdf):
    aoi_gdf = aoi_gdf.to_crs(4326)

    # Add a time_column to the gdf
    aoi_gdf["time"] = pd.Timestamp.utcnow() - pd.Timedelta(days=365)

    img_coll = (
        ee.ImageCollection("MODIS/061/MYD13A1")
        .select("NDVI")
        .map(
            lambda img: img.multiply(0.0001)
            .set("system:time_start", img.get("system:time_start"))
            .set("id", img.get("id"))
        )
        .sort("system:time_start")
    )

    params = {
        "time_col_name": "time",
        "stack_limit_before": 10,
        "stack_limit_after": 10,
        "img_coll": img_coll,
        "region_reducer": "mean",
        "scale": 500.0,
    }

    results = ecoscope.io.eetools.chunk_gdf(
        gdf=aoi_gdf,
        label_func=ecoscope.io.eetools.label_gdf_with_temporal_image_collection_by_feature,
        label_func_kwargs=params,
        df_chunk_size=10,
        max_workers=1,
    )

    assert results["NDVI"].explode().mean() > 0


def test_label_gdf_with_temporal_image_collection_by_features_relocations(movbank_relocations):
    tmp_gdf = movbank_relocations[["fixtime", "geometry"]].iloc[0:1000]

    img_coll = ee.ImageCollection("MODIS/MCD43A4_006_NDVI").select("NDVI")  # Daily NDVI images

    params = {
        "time_col_name": "fixtime",
        "stack_limit_before": 1,
        "stack_limit_after": 1,
        "img_coll": img_coll,
        "region_reducer": "toList",
        "scale": 1.0,
    }

    results = ecoscope.io.eetools.chunk_gdf(
        gdf=tmp_gdf,
        label_func=ecoscope.io.eetools.label_gdf_with_temporal_image_collection_by_feature,
        label_func_kwargs=params,
        df_chunk_size=25000,
        max_workers=1,
    )

    assert results["NDVI"].explode().mean() > 0


def test_label_gdf_with_img(movbank_relocations):
    tmp_gdf = movbank_relocations[["geometry"]]
    tmp_gdf = tmp_gdf[0:1000]

    img = ee.Image("USGS/SRTMGL1_003").select("elevation")

    params = {"img": img, "region_reducer": "toList", "scale": 1.0}

    results = ecoscope.io.eetools.chunk_gdf(
        gdf=tmp_gdf,
        label_func=ecoscope.io.eetools.label_gdf_with_img,
        label_func_kwargs=params,
        df_chunk_size=20000,
        max_workers=1,
    )

    assert results["list"].mean() > 0


# a subset to ensure we're checking exact match and nearest cases
# includes 3 timestamps, midnight, am, pm
@pytest.fixture
def movbank_relocations_fixed_subset(movbank_relocations):
    return movbank_relocations.loc[329730794:329730795]._append(movbank_relocations.loc[329730810])


@pytest.mark.parametrize(
    "n_before, n_after, output_list",
    [
        (0, 0, [["2009_01_29"], ["2009_01_29"], ["2009_01_30"]]),
        (1, 0, [["2009_01_28", "2009_01_29"], ["2009_01_29"], ["2009_01_29"]]),
        (0, 1, [["2009_01_29", "2009_01_30"], ["2009_01_30"], ["2009_01_30"]]),
        (
            2,
            2,
            [
                ["2009_01_27", "2009_01_28", "2009_01_29", "2009_01_30", "2009_01_31"],
                ["2009_01_28", "2009_01_29", "2009_01_30", "2009_01_31"],
                ["2009_01_28", "2009_01_29", "2009_01_30", "2009_01_31"],
            ],
        ),
    ],
)
def test_match_gdf_to_img_coll_ids_by_image_count(movbank_relocations_fixed_subset, n_before, n_after, output_list):
    results = ecoscope.io.eetools._match_gdf_to_img_coll_ids(
        gdf=movbank_relocations_fixed_subset,
        time_col="fixtime",
        img_coll=ee.ImageCollection("MODIS/MCD43A4_006_NDVI").select("NDVI"),
        output_col_name="img_ids",
        n_before=n_before,
        n_after=n_after,
        n="images",
    )["img_ids"].to_list()

    # midnight - exact match case
    assert results[0] == output_list[0]
    # am
    assert results[1] == output_list[1]
    # pm
    assert results[2] == output_list[2]


@pytest.mark.parametrize(
    "n_before, n_after, output_list",
    [
        (0, 0, [["2009_01_29"], [], []]),
        (1, 0, [["2009_01_28", "2009_01_29"], ["2009_01_29"], ["2009_01_29"]]),
        (0, 1, [["2009_01_29", "2009_01_30"], ["2009_01_30"], ["2009_01_30"]]),
        (
            2,
            2,
            [
                ["2009_01_27", "2009_01_28", "2009_01_29", "2009_01_30", "2009_01_31"],
                ["2009_01_28", "2009_01_29", "2009_01_30", "2009_01_31"],
                ["2009_01_28", "2009_01_29", "2009_01_30", "2009_01_31"],
            ],
        ),
    ],
)
def test_match_gdf_to_img_coll_ids_by_day(movbank_relocations_fixed_subset, n_before, n_after, output_list):
    output = ecoscope.io.eetools._match_gdf_to_img_coll_ids(
        gdf=movbank_relocations_fixed_subset,
        time_col="fixtime",
        img_coll=ee.ImageCollection("MODIS/MCD43A4_006_NDVI").select("NDVI"),
        output_col_name="img_ids",
        n_before=n_before,
        n_after=n_after,
        n="days",
    )

    print(output[["fixtime", "img_ids"]])
    results = output["img_ids"].to_list()

    # midnight - exact match case
    assert results[0] == output_list[0]
    # am
    assert results[1] == output_list[1]
    # pm
    assert results[2] == output_list[2]



import warnings

import astroplan
import astropy.coordinates
import astropy.time
import numpy as np
import pyproj


def to_EarthLocation(geometry):
    """
    Location on Earth, initialized from geocentric coordinates.

    Parameters
    ----------
    geometry: Geometry
        GeoDataFrame's geometry column

    Returns
    -------
    astropy.coordinates.EarthLocation.
    """
    geometry = geometry.to_crs(4326)
    trans = pyproj.Transformer.from_proj(
        proj_from="+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs",
        proj_to="+proj=geocent +ellps=WGS84 +datum=WGS84 +units=m +no_defs",
    )
    return astropy.coordinates.EarthLocation.from_geocentric(
        *trans.transform(xx=geometry.x, yy=geometry.y, zz=np.zeros(geometry.shape[0])), unit="m"
    )


def is_night(geometry, time):
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", "Geometry is in a geographic CRS.", UserWarning)
        return astroplan.Observer(to_EarthLocation(geometry.centroid)).is_night(time)



from ecoscope import analysis, base, contrib, io, mapping, plotting

ASCII = """\
 _____
|   __|___ ___ ___ ___ ___ ___ ___
|   __|  _| . |_ -|  _| . | . | -_|
|_____|___|___|___|___|___|  _|___|
                          |_|
"""

__initialized = False


def init(silent=False, selenium=False, force=False):
    """
    Initializes the environment with ecoscope-specific customizations.

    Parameters
    ----------
    silent : bool, optional
        Removes console output
    selenium : bool, optional
        Installs selenium webdriver in a colab environment
    force : bool, optional
        Ignores `__initialized`

    """

    global __initialized
    if __initialized and not force:
        if not silent:
            print("Ecoscope already initialized.")
        return

    import pandas as pd

    pd.options.plotting.backend = "plotly"

    from tqdm.auto import tqdm

    tqdm.pandas()

    import warnings

    from shapely.errors import ShapelyDeprecationWarning

    warnings.filterwarnings(action="ignore", category=ShapelyDeprecationWarning)
    warnings.filterwarnings(action="ignore", category=FutureWarning)
    warnings.filterwarnings("ignore", message=".*initial implementation of Parquet.*")

    import geopandas as gpd

    def explore(data, *args, **kwargs):
        """
        Monkey-patched `geopandas.explore._explore` to use EcoMap instead.
        """
        initialized = "m" in kwargs
        if not initialized:
            kwargs["m"] = mapping.EcoMap()

        if isinstance(kwargs["m"], mapping.EcoMap):
            m = kwargs.pop("m")
            m.add_gdf(data, *args, **kwargs)
            if not initialized:
                m.zoom_to_bounds(data.geometry.to_crs(4326).total_bounds)
            return m
        else:
            return gpd.explore._explore(data, *args, **kwargs)

    gpd.GeoDataFrame.explore = explore
    gpd.GeoSeries.explore = explore

    import plotly.io as pio

    pio.templates.default = "seaborn"

    import sys

    if "google.colab" in sys.modules and selenium:
        from IPython import get_ipython

        shell_text = """\
cat > /etc/apt/sources.list.d/debian.list <<'EOF'
deb [arch=amd64 signed-by=/usr/share/keyrings/debian-bookworm.gpg] http://deb.debian.org/debian bookworm main
deb [arch=amd64 signed-by=/usr/share/keyrings/debian-bookworm-updates.gpg]\
        http://deb.debian.org/debian bookworm-updates main
deb [arch=amd64 signed-by=/usr/share/keyrings/debian-security-bookworm.gpg]\
        http://deb.debian.org/debian-security bookworm/updates main
EOF

apt-key adv --keyserver keyserver.ubuntu.com --recv-keys DCC9EFBF77E11517
apt-key adv --keyserver keyserver.ubuntu.com --recv-keys 648ACFD622F3D138
apt-key adv --keyserver keyserver.ubuntu.com --recv-keys 112695A0E562B32A

apt-key export 77E11517 | gpg --dearmour -o /usr/share/keyrings/debian-bookworm.gpg
apt-key export 22F3D138 | gpg --dearmour -o /usr/share/keyrings/debian-bookworm-updates.gpg
apt-key export E562B32A | gpg --dearmour -o /usr/share/keyrings/debian-security-bookworm.gpg

cat > /etc/apt/preferences.d/chromium.pref << 'EOF'
Package: *
Pin: release a=eoan
Pin-Priority: 500


Package: *
Pin: origin "deb.debian.org"
Pin-Priority: 300


Package: chromium*
Pin: origin "deb.debian.org"
Pin-Priority: 700
EOF

apt-get update
apt-get install chromium chromium-driver

pip install selenium
"""

        if silent:
            from IPython.utils import io

            with io.capture_output():
                get_ipython().run_cell_magic("shell", "", shell_text)
        else:
            get_ipython().run_cell_magic("shell", "", shell_text)

    __initialized = True
    if not silent:
        print(ASCII)


__all__ = ["analysis", "base", "contrib", "init", "io", "mapping", "plotting"]



import geopandas as gpd
import geopandas.testing
import pandas as pd
from shapely.geometry import MultiLineString, Polygon

import ecoscope
from ecoscope.analysis import geofence


def test_geofence_crossing():
    obs = gpd.read_file("tests/sample_data/vector/observations.geojson")
    obs["recorded_at"] = pd.to_datetime(obs["recorded_at"], utc=True)
    relocations = ecoscope.base.Relocations.from_gdf(obs, groupby_col="source_id", time_col="recorded_at")
    trajectory = ecoscope.base.Trajectory.from_relocations(relocations)

    region = Polygon(
        [
            [18.45703125, 4.609278084409835],
            [18.6767578125, 4.521666342614804],
            [18.6328125, 4.653079918274051],
            [17.4462890625, -1.1425024037061522],
            [36.6943359375, -4.653079918274038],
            [41.0888671875, -1.9332268264771106],
            [41.66015625, 4.039617826768437],
            [18.45703125, 4.609278084409835],
        ]
    )

    equator_fence = MultiLineString([((0, 0), (90, 0)), ((180, 0), (270, 0))])

    geofences = [
        geofence.GeoFence(
            geometry=equator_fence,
            unique_id="f81bff6a-9906-4a22-88cc-aa7ab958d876",
            fence_name="equator_fence",
            warn_level="warning",
        )
    ]

    regions = [
        geofence.Region(
            region_name="East & Central Africa",
            geometry=region,
            unique_id="1c24d330-c5df-4fd6-aa78-1e07f7b14edd",
        )
    ]

    gf_profile = geofence.GeoCrossingProfile(geofences=geofences, regions=regions)
    df = geofence.GeoFenceCrossing.analyse(geocrossing_profile=gf_profile, trajectory=trajectory)
    geofence_crossing_point = gpd.read_feather("tests/test_output/geofence_crossing_point.feather")
    gpd.testing.assert_geodataframe_equal(df, geofence_crossing_point)
    assert df.crs == geofence_crossing_point.crs



import collections
import dataclasses
import typing

import geopandas as gpd
import pandas as pd
import shapely

import ecoscope


class Region(collections.UserDict):
    def __init__(self, geometry: typing.Any, unique_id: str = None, region_name: str = None):
        super().__init__(
            geometry=geometry,
            unique_id=unique_id,
            region_name=region_name,
        )


class GeoFence(collections.UserDict):
    """Represents Virtual Fence Boundary"""

    def __init__(
        self,
        geometry: typing.Any,
        unique_id: str = None,
        fence_name: str = None,
        warn_level: str = None,
    ):
        super().__init__(
            geometry=geometry,
            unique_id=unique_id,
            fence_name=fence_name,
            warn_level=warn_level,
        )


@dataclasses.dataclass
class GeoCrossingProfile:
    geofences: typing.List[GeoFence]
    regions: typing.List[Region]

    @property
    def geofence_df(self):
        return gpd.GeoDataFrame(self.geofences, crs=4326)

    @property
    def region_df(self):
        return gpd.GeoDataFrame(self.regions, crs=4326)


class GeoFenceCrossing:
    @classmethod
    def analyse(
        cls,
        geocrossing_profile: GeoCrossingProfile,
        trajectory: ecoscope.base.Trajectory,
    ):
        """
        Analyze the trajectory of each subject in relation to set of virtual fences and regions to determine where/when
        the polylines were crossed and what the containment of the individual was before and
        after any geofence crossings.

        Parameters
        ----------
        geocrossing_profile: GeoCrossingProfile
            Object that contains the geonfences and regions
        trajectory: ecoscope.base.Trajectory
            Geodataframe stores goemetry, speed_kmhr, heading etc. for each subject.

        Returns
        -------
            ecoscope.base.EcoDataFrame

        """
        trajectory = trajectory.copy()
        trajectory["start_point"] = shapely.get_point(trajectory.geometry, 0)
        trajectory["end_point"] = shapely.get_point(trajectory.geometry, 1)

        def apply_func(fence):
            geofence = fence.geometry
            traj = trajectory.loc[trajectory.intersects(geofence)].copy()
            traj["segment_geometry"] = traj["geometry"]
            traj.set_geometry(traj.intersection(geofence), inplace=True)
            traj = traj.explode(index_parts=False)
            traj = traj.loc[traj["geometry"].type == "Point"]

            dist = traj.segment_geometry.project(traj.geometry, normalized=True)
            traj["crossing_time"] = dist * (traj.segment_end - traj.segment_start) + traj.segment_start

            traj["unique_id"] = fence.unique_id
            traj["fence_name"] = fence.fence_name
            traj["warn_level"] = fence.warn_level

            # Determine containment of the subject before and after the crossing
            for index, colname in enumerate(["start_region_ids", "end_region_ids"]):
                traj[colname] = (
                    geocrossing_profile.region_df.sjoin(
                        gpd.GeoDataFrame(
                            geometry=shapely.get_point(traj["segment_geometry"], index),
                            index=traj.index,
                            crs=4326,
                        ),
                        predicate="contains",
                    )
                    .set_index("index_right")["unique_id"]
                    .groupby(level=0)
                    .agg(list)
                )
            return traj

        fences = geocrossing_profile.geofence_df
        df = pd.concat([apply_func(fence) for _, fence in fences.iterrows()])
        df.drop(["start_point", "end_point"], axis=1, inplace=True)
        return ecoscope.base.EcoDataFrame(df, geometry="geometry")



import base64
import os
import tempfile
import time
import typing
import urllib
import warnings


import ee
import folium
import geopandas as gpd
import matplotlib as mpl
import numpy as np
import pandas as pd
import datashader as ds
import rasterio
import selenium.webdriver
from branca.colormap import StepColormap
from branca.element import MacroElement, Template

import ecoscope
from ecoscope.contrib.foliumap import Map

warnings.filterwarnings("ignore", "GeoSeries.isna", UserWarning)


class EcoMapMixin:
    def add_speedmap(
        self,
        trajectory: gpd.GeoDataFrame,
        classification_method: str = "equal_interval",
        num_classes: int = 6,
        speed_colors: typing.List = None,
        bins: typing.List = None,
        legend: bool = True,
    ):

        speed_df = ecoscope.analysis.SpeedDataFrame.from_trajectory(
            trajectory=trajectory,
            classification_method=classification_method,
            num_classes=num_classes,
            speed_colors=speed_colors,
            bins=bins,
        )
        self.add_gdf(speed_df, color=speed_df["speed_colour"])

        if legend:
            self.add_legend(legend_dict=dict(zip(speed_df.label, speed_df.speed_colour)))

        return speed_df


class EcoMap(EcoMapMixin, Map):
    def __init__(self, *args, static=False, print_control=True, **kwargs):
        kwargs["attr"] = kwargs.get("attr", " ")
        kwargs["canvas"] = kwargs.get("canvas", True)
        kwargs["control_scale"] = kwargs.get("control_scale", False)
        kwargs["height"] = kwargs.get("height", 600)
        kwargs["width"] = kwargs.get("width", 800)

        if static:
            print_control = False
            kwargs["draw_control"] = kwargs.get("draw_control", False)
            kwargs["fullscreen_control"] = kwargs.get("fullscreen_control", False)
            kwargs["layers_control"] = kwargs.get("layers_control", False)
            kwargs["measure_control"] = kwargs.get("measure_control", False)
            kwargs["zoom_control"] = kwargs.get("zoom_control", False)
            kwargs["search_control"] = kwargs.get("search_control", False)

        self.px_height = kwargs["height"]
        self.px_width = kwargs["width"]

        super().__init__(*args, **kwargs)

        if print_control:
            self.add_print_control()

    def add_gdf(self, data, *args, simplify_tolerance=None, **kwargs):
        """
        Wrapper for `geopandas.explore._explore`.
        """

        data = data.copy()
        data = data.to_crs(4326)
        data = data.loc[(~data.geometry.isna()) & (~data.geometry.is_empty)]

        if simplify_tolerance is not None:
            data = data.simplify(simplify_tolerance)

        if isinstance(data, gpd.GeoDataFrame):
            for col in data:
                if pd.api.types.is_datetime64_any_dtype(data[col]):
                    data[col] = data[col].astype("string")

        kwargs["m"] = self
        kwargs["tooltip"] = kwargs.get("tooltip", False)
        gpd.explore._explore(data, *args, **kwargs)

    def add_legend(self, *args, **kwargs):
        """
        Patched method for allowing legend hex colors to start with a "#".
        """
        legend_dict = kwargs.get("legend_dict")
        if legend_dict is not None:
            kwargs["legend_dict"] = {
                k: v[1:] if isinstance(v, str) and v.startswith("#") else v for k, v in legend_dict.items()
            }

        return super().add_legend(*args, **kwargs)

    def add_north_arrow(self, imagePath="", position="topright", angle=0, scale=1.0):
        """
        Parameters
        ----------
        imagePath : str
            Path to the image file for the north arrow. If not provided, default SVG image is uploaded.
        position : str
            Possible values are 'topleft', 'topright', 'bottomleft' or 'bottomright'.
        angle : int
            Angle for the north arrow to be rotated.
        scale : float
            Scale dimensions of north arrow.

        """

        self.add_child(
            NorthArrowElement(
                html=f"""\
            <svg width="{100*scale}px" height="{100*scale}px" version="1.1" viewBox="0 0 773 798" xmlns="http://www.w3.org/2000/svg">
<path transform="translate(0 798) scale(1 -1)" d="m674 403-161 48q-17 48-66 70l-46 166-46-167q-22-9-38-25t-29-45l-159-47 159-49q15-44 67-68l46-164 48 164q39 17 64 69zm-163 0q0-49-32-81-33-34-78-34-46 0-77 34-31 31-31 81 0 46 31 80t77 34q45 0 78-34 32-34 32-80zm-12 1q-5 7-7.5 17.5t-4 21.5-4.5 21-9 16q-7 6-17 9.5t-20.5 6-20 6-15.5 9.5v-107h98zm-98-108v108h-99l3-3 23-75q6-6 16.5-9.5t21-5.5 20-5.5 15.5-9.5zm-280 152h-26v-2q5 0 6-1 3-3 3-6 0-2-0.5-4t-1.5-7l-18-48-16 47q-3 9-3 12 0 7 7 7h2v2h-34v-2q2 0 3-1l3-3q2 0 2-2 2-1 4-5l5-15-12-42-17 50q-3 9-3 11 0 7 6 7h2v2h-33v-2q8 0 10-6 1-2 3-9l27-74h5l15 53 19-53h2l27 71q2 10 3 11 5 7 10 7v2zm325 350h-29v-3q7 0 10-4 1-1 1-11v-35l-42 53h-32v-3q7-2 12-6l2-3v-62q0-13-12-13v-2h29v2h-2q-4 0-7 2.5t-3 10.5v55l58-72h3v73q0 9 1 10.5t8 3.5l3 1v3zm207-395h-130q0 16-6 42zm-212-119-40-141v135q9 0 19 1t21 5zm-154 78-137 41h130q0-10 2-19.5t5-21.5zm114 168q-25 0-39-8l39 142v-134zm372-148h-3q-3-4-5-7.5t-4-5.5q-5-5-17-5h-19q-3 0-3 5v35h20q8 0 10-6 1-1 1-3 0-3 1-4h3v30h-3q-2-9-4-11t-8-2h-20v35h24q7 0 8-1 4-1 9-14h3l-1 20h-69v-2h3q7 0 8-4 2-2 2-9v-58q0-11-4-12-1-1-6-1h-3v-3h68zm-340-358q0 9-5.5 14.5t-20.5 14.5q-9 5-13 9l-5 5q-3 10-3 7 0 14 14 14 18 0 24-26h2v31h-2q-2-6-5-6-4 0-5 1-8 5-15 5-11 0-17.5-7t-6.5-17q0-13 9-19 6-4 16.5-10.5t12.5-8.5q8-7 8-13 0-14-18-14-13 0-18 5.5t-7 20.5h-2v-30h2q0 5 3 5l16-5h8q12 0 20 7t8 17z"/>
            </svg>""",  # noqa
                imagePath=imagePath,
                position=position,
                angle=angle,
                scale=scale,
            )
        )

    def add_scale_bar(self, style="single", position="bottomleft", imperial=False, force=False):
        """
        Parameters
        ----------
        style : str, optional
            Possible values are 'single', 'double', 'empty' or 'comb'. Default is 'single'.
        position : str, optional
            Possible values are 'topleft', 'topright', 'bottomleft' or 'bottomright'.
        imperial : bool, optional
            If True, the scale bar uses miles and feet. Default is False.
        force : bool, optional
            If True, adds index=0 to the add_child() call, forcing the scale bar 'behind' Folium's default controls
        """

        svg = (
            """<svg xmlns="http://www.w3.org/2000/svg" id="test" style="width:300;height:40px;"><rect id="border" style="stroke:#000;fill:#FFF;" height="40%" width="75%" x="5%" y="2%"/><rect id="first_block" style="fill:#000" height="20%" width="37.5%" x="5%" y="2%"/><rect id="second_block" style="fill:#000" height="20%" width="37.5%" x="42.5%" y="22%"/><text id="zero" text-anchor="middle" font-size="20" x="5%" y="95%">0</text><text id="half_scale" font-size="20" text-anchor="middle" x="42.5%" y="95%">15</text><text id="scale" font-size="20" text-anchor="middle" x="80%" y="95%">30</text><text id="unit" font-size="20" x="82%" y="42%">km</text></svg>"""  # noqa
            if style == "double"
            else """<svg xmlns="http://www.w3.org/2000/svg" id="test" style="width:300;height:40px;"><rect id="border" style="stroke:#000;fill:#FFF;" height="50%" width="75%" x="5%" y="48%" /><text id="zero" font-size="20" x="0%" y="95%" visibility="hidden">0</text><text id="half_scale" font-size="20" text-anchor="middle" x="42.5%" y="95%" visibility="hidden">15</text><text id="scale" font-size="20" text-anchor="middle" x="42.5%" y="92%">30</text><text id="unit" font-size="20" x="82%" y="95%">km</text></svg>"""  # noqa
            if style == "empty"
            else """<svg xmlns="http://www.w3.org/2000/svg" id="test" style="width:300;height:40px;"><svg y="50%" x="5%" width="75%" height="48%" viewBox="-0.4 0 101 101" preserveAspectRatio="none"><path d="M 0 50 L 0 99 h 50 V 50 V 99 H 100 V 50" vector-effect="non-scaling-stroke" fill="None" stroke-width="2" stroke="#000"/></svg><text id="zero" font-size="20" text-anchor="middle" x="5%" y="50%">0</text><text id="half_scale" font-size="20" text-anchor="middle" y="50%" x="42.5%">15</text><text id="scale" font-size="20" text-anchor="middle" y="50%" x="80%">1500</text><text id="unit" font-size="20" x="85%" y="98%">km</text></svg>"""  # noqa
            if style == "comb"
            else """<svg xmlns="http://www.w3.org/2000/svg" id="test" style="width:300;height:40px;"><rect id="border" style="stroke:#000;fill:#FFF;" height="40%" width="75%" x="5%" y="2%"/><rect id="first_block" style="fill:#000" height="40%" width="37.5%" x="5%" y="2%"/><text id="zero" text-anchor="middle" font-size="20" x="5%" y="95%">0</text><text id="half_scale" font-size="20" text-anchor="middle" x="42.5%" y="95%">15</text><text id="scale" font-size="20" text-anchor="middle" x="80%" y="95%">30</text><text id="unit" font-size="20" x="82%" y="42%">km</text></svg>"""  # noqa
        )

        self.add_child(ScaleElement(svg, position=position, imperial=imperial), index=0 if force else None)

    def add_title(
        self,
        title: str,
        font_size: str = "32px",
        font_style: str = "normal",
        font_family: typing.Union[str, list] = "Helvetica",
        font_color: typing.Union[str, tuple] = "rgba(0,0,0,1)",
        position: dict = None,
        background_color: typing.Union[str, tuple] = "#FFFFFF99",
        outline: typing.Union[str, dict] = "0px solid rgba(0, 0, 0, 0)",
        **kwargs,
    ):
        """
        Parameters
        ----------
        title : str
            Text of title.
        font_size : str
            CSS font size that includes units.
            font_style : str
        font_family:str"Helvetica",
            Font family selection; Could be one or more separated by spaces.
        font_color : str
            Text color (css color property); supports rgb, rgba and hex string formats.
        position : dict|None
            Dict object with top, left and bottom margin values for the title container position.
            ex. {
                    "top": "10px",
                    "left": "25px",
                    "right": "0px",
                    "bottom": "0px"
                }
            All keys are optional in the dictionary (could be passed some of them as necessary).
            Values could be passed as px or accepted css metric. Default None.
        background_color : str
            Box background color; supports rgb, rgba and hex string formats. Default '#FFFFFF99'.
        outline : str
            Element outline values (width style color_with_transparency).
            Could be passed as a string with spaced separated values or a dict structure:
            ex. {
                    "width": "1px",
                    "style": "solid",
                    "color": "#FFFFFF99" # or rgb/rgba tuple (0, 0, 0, 0)
                }
        kwargs
            Additional style kwargs. Underscores in keys will be replaced with dashes
        """
        position_styles = ""
        if isinstance(position, dict):
            VALID_POSITION_OPTIONS = ["top", "bottom", "left", "right"]
            position_styles = " ".join([f"{k}:{v};" for k, v in position.items() if k in VALID_POSITION_OPTIONS])
        else:
            position_styles = "left: 50%;"
        if isinstance(font_family, str):
            font = font_family
        elif isinstance(font_family, list):
            font = " ".join(font_family)
        if isinstance(font_color, tuple):
            if font_color.__len__() == 3:
                fc = "rgb({},{},{})".format(font_color[0], font_color[1], font_color[2])
            elif font_color.__len__() == 4:
                fc = "rgba({},{},{},{})".format(font_color[0], font_color[1], font_color[2], font_color[3])
        elif isinstance(font_color, str):
            fc = font_color if font_color.startswith("#") else f"#{font_color}"
        if isinstance(background_color, tuple):
            if background_color.__len__() == 3:
                bkg = "rgb({},{},{})".format(background_color[0], background_color[1], background_color[2])
            elif background_color.__len__() == 4:
                bkg = "rgba({},{},{},{})".format(
                    background_color[0], background_color[1], background_color[2], background_color[3]
                )
        elif isinstance(background_color, str):
            bkg = background_color if background_color.startswith("#") else f"#{background_color}"
        outline = (
            outline
            if isinstance(outline, str)
            else "{} {} {}".format(outline["width"], outline["style"], outline["color"])
        )
        title_html = f"""\
        <div style="position: absolute; {position_styles}">
            <p style="position: relative;
                    left: -50%;
                    border: 1px solid #000;
                    border-radius: 5px;
                    background-color: {bkg};
                    padding: 3px;
                    outline: {outline};
                    font-size: { font_size };
                    font-family: {font};
                    font-style:{font_style};
                    color: {fc};
                    { "".join([f"{k.replace('_', '-')}: {v};" for k, v in kwargs.items() ]) }
                    ">
            { title }
            </p>
        </div>
        """
        self.add_child(FloatElement(title_html, top=0, left=0, right=0))

    def _repr_html_(self, **kwargs):
        if kwargs.get("fill_parent", False):
            original_width = self._parent.width
            original_height = self._parent.height

            self._parent.width = "100%"
            self._parent.height = "100%"

        html = (
            super()
            ._repr_html_(**kwargs)
            .replace(urllib.parse.quote("crs: L.CRS."), urllib.parse.quote("attributionControl: false, crs: L.CRS."))
        )

        # this covers the (probably) rare case where someone in a Jupyter setting:
        # creates a map, calls to_html(), continues changing the map, and then displays via _repr_html_()
        if kwargs.get("fill_parent", False):
            self._parent.width = original_width
            self._parent.height = original_height

        return html

    def to_html(self, outfile, fill_parent=True, **kwargs):
        """
        Parameters
        ----------
        outfile : str, Pathlike
            Output destination

        """
        with open(outfile, "w") as file:
            file.write(self._repr_html_(fill_parent=fill_parent, **kwargs))

    def to_png(self, outfile, sleep_time=5, **kwargs):
        """
        Parameters
        ----------
        outfile : str, Pathlike
            Output destination
        sleep_time : int, optional
            Additional seconds to wait before taking screenshot. Should be increased if map tiles in the output haven't
            fully loaded but can also be decreased in most cases.

        """

        with tempfile.NamedTemporaryFile(mode="w", suffix=".html") as tmp:
            super().to_html(tmp.name, **kwargs)
            chrome_options = selenium.webdriver.chrome.options.Options()
            chrome_options.add_argument("--headless")
            chrome_options.add_argument("--no-sandbox")
            chrome_options.add_argument("--disable-dev-shm-usage")
            driver = selenium.webdriver.Chrome(options=chrome_options)
            if self.px_width and self.px_height:
                driver.set_window_size(width=self.px_width, height=self.px_height)
            driver.get(f"file://{os.path.abspath(tmp.name)}")
            time.sleep(sleep_time)
            driver.save_screenshot(outfile)

    def add_ee_layer(self, ee_object, visualization_params, name) -> None:
        """
        Method for displaying Earth Engine image tiles.

        Parameters
        ----------
        ee_object
        visualization_params
        name

        Returns
        -------
        None

        """

        try:
            if isinstance(ee_object, ee.image.Image):
                map_id_dict = ee.Image(ee_object).getMapId(visualization_params)
                folium.raster_layers.TileLayer(
                    tiles=map_id_dict["tile_fetcher"].url_format,
                    attr="Google Earth Engine",
                    name=name,
                    overlay=True,
                    control=True,
                ).add_to(self)

            elif isinstance(ee_object, ee.imagecollection.ImageCollection):
                ee_object_new = ee_object.mosaic()
                map_id_dict = ee.Image(ee_object_new).getMapId(visualization_params)
                folium.raster_layers.TileLayer(
                    tiles=map_id_dict["tile_fetcher"].url_format,
                    attr="Google Earth Engine",
                    name=name,
                    overlay=True,
                    control=True,
                ).add_to(self)

            elif isinstance(ee_object, ee.geometry.Geometry):
                folium.GeoJson(data=ee_object.getInfo(), name=name, overlay=True, control=True).add_to(self)

            elif isinstance(ee_object, ee.featurecollection.FeatureCollection):
                ee_object_new = ee.Image().paint(ee_object, 0, 2)
                map_id_dict = ee.Image(ee_object_new).getMapId(visualization_params)
                folium.raster_layers.TileLayer(
                    tiles=map_id_dict["tile_fetcher"].url_format,
                    attr="Google Earth Engine",
                    name=name,
                    overlay=True,
                    control=True,
                ).add_to(self)

        except Exception as exc:
            print(f"{exc}. Could not display {name}")

    def zoom_to_bounds(self, bounds):
        """Zooms to a bounding box in the form of [minx, miny, maxx, maxy].

        Parameters
        ----------
        bounds : [x1, y1, x2, y2]
            Map extent in WGS 84
        """

        assert -180 < bounds[0] <= bounds[2] < 180
        assert -90 < bounds[1] <= bounds[3] < 90

        self.fit_bounds([[bounds[1], bounds[0]], [bounds[3], bounds[2]]])

    def zoom_to_gdf(self, gs):
        """
        Parameters
        ----------
        gs : gpd.GeoSeries or gpd.GeoDataFrame
            Geometry to adjust map bounds to. CRS will be converted to WGS 84.

        """

        self.zoom_to_bounds(gs.geometry.to_crs(4326).total_bounds)

    def add_local_geotiff(self, path, zoom=False, cmap=None, colorbar=True):
        """
        Displays a local GeoTIFF.

        Parameters
        ----------
        path : str, Pathlike
            Path to local GeoTIFF
        zoom : bool
            Zoom to displayed GeoTIFF
        cmap : `matplotlib.colors.Colormap` or str or None
            Matplotlib colormap to apply to raster
        colorbar : bool
            Whether to add colorbar for provided `cmap`. Does nothing if `cmap` is not provided.
        """

        with rasterio.open(path) as src:
            transform, width, height = rasterio.warp.calculate_default_transform(
                src.crs, "EPSG:4326", src.width, src.height, *src.bounds
            )
            kwargs = src.meta.copy()
            kwargs.update({"crs": "EPSG:4326", "transform": transform, "width": width, "height": height})

            if cmap is None:
                im = [rasterio.band(src, i + 1) for i in range(src.count)]
            else:
                cmap = mpl.cm.get_cmap(cmap)
                kwargs["count"] = 4
                im = rasterio.band(src, 1)[0].read()[0]
                im_min, im_max = np.nanmin(im), np.nanmax(im)
                im = np.rollaxis(cmap((im - im_min) / (im_max - im_min), bytes=True), -1)
                if colorbar:
                    if isinstance(im_min, np.integer) and im_max - im_min < 256:
                        self.add_child(
                            StepColormap(
                                [mpl.colors.rgb2hex(color) for color in cmap(np.linspace(0, 1, 1 + im_max - im_min))],
                                index=np.arange(1 + im_max - im_min),
                                vmin=im_min,
                                vmax=im_max + 1,
                            )
                        )
                    else:
                        self.add_child(
                            StepColormap(
                                [mpl.colors.rgb2hex(color) for color in cmap(np.linspace(0, 1, 256))],
                                vmin=im_min,
                                vmax=im_max,
                            )
                        )

            with rasterio.io.MemoryFile() as memfile:
                with memfile.open(**kwargs) as dst:
                    for i in range(kwargs["count"]):
                        rasterio.warp.reproject(
                            source=im[i],
                            destination=rasterio.band(dst, i + 1),
                            src_transform=src.transform,
                            src_crs=src.crs,
                            dst_transform=transform,
                            dst_crs="EPSG:4326",
                            resampling=rasterio.warp.Resampling.nearest,
                        )

                url = "data:image/tiff;base64," + base64.b64encode(memfile.read()).decode("utf-8")

        self.add_child(GeoTIFFElement(url, zoom))

    def add_print_control(self):
        self.add_child(PrintControl())

    def add_datashader_gdf(
        self,
        gdf,
        geom_type,
        name=None,
        width=600,
        height=600,
        cmap=["lightblue", "darkblue"],
        ds_agg=None,
        zoom=True,
        opacity=1,
        **kwargs,
    ):
        """
        Overlays a static visualization of the given gdf generated using Datashader

        Parameters
        ----------
        gdf : geopandas.GeoDataFrame
            GeoDataFrame used to create visualization (geometry must be projected to a CRS)
        geom_type : str
            The Datashader canvas() function to use valid values are 'polygon', 'line', 'point'
        name : string
            The name of the image layer to be displayed in Folium layer control
        width : int
            The canvas width in pixels, determines the resolution of the generated image
        height : int
            The canvas height in pixels, determines the resolution of the generated image
        cmap : list of colors or matplotlib.colors.Colormap
            The colormap to use for the generated image
        ds_agg : datashader.reductions function
            The Datashader reduction to use
        zoom : bool
            Zoom to the generated image
        opacity : float
            Sets opacity of overlaid image
        kwargs
            Additional kwargs passed to datashader.transfer_functions.shade
        """

        gdf = gdf.to_crs(epsg=4326)
        bounds = gdf.geometry.total_bounds
        canvas = ds.Canvas(width, height)

        if geom_type == "polygon":
            func = canvas.polygons
        elif geom_type == "line":
            func = canvas.line
        elif geom_type == "point":
            func = canvas.points
        else:
            raise ValueError("geom_type must be 'polygon', 'line' or 'point'")

        agg = func(gdf, geometry="geometry", agg=ds_agg)
        img = ds.tf.shade(agg, cmap, **kwargs)

        folium.raster_layers.ImageOverlay(
            image=img.to_pil(),
            bounds=[[bounds[1], bounds[0]], [bounds[3], bounds[2]]],
            opacity=opacity,
            mercator_project=True,
            name=name,
        ).add_to(self)

        if zoom:
            self.zoom_to_gdf(gdf)


class ControlElement(MacroElement):
    """
    Class to wrap arbitrary HTML as Leaflet Control.

    Parameters
    ----------
    html : str
        HTML to render an element from.
    position : str
        Possible values are 'topleft', 'topright', 'bottomleft' or 'bottomright'.

    """

    _template = Template(
        """
        {% macro script(this, kwargs) %}
        var {{ this.get_name() }} = L.Control.extend({
            onAdd: function(map) {
                var template = document.createElement('template');
                template.innerHTML = `{{ this.html }}`.trim();
                return template.content.firstChild;
            }
        });
        (new {{ this.get_name() }}({{ this.options|tojson }})).addTo({{this._parent.get_name()}});

        {% endmacro %}
    """
    )

    def __init__(self, html, position="bottomright"):
        super().__init__()
        self.html = html
        self.options = folium.utilities.parse_options(
            position=position,
        )


class NorthArrowElement(ControlElement):
    """
    Class to wrap arbitrary HTML or PNG as Leaflet Control.

    Parameters
    ----------
    html : str
        HTML to render an element from.
    imagePath: str
        Path to the image file for the north arrow. If not provided, default SVG image is uploaded.
    position : str
        Possible values are 'topleft', 'topright', 'bottomleft' or 'bottomright'.
    angle : int
        Angle for the north arrow to be rotated.
    scale : float
        Scale dimensions of north arrow.

    """

    _template = Template(
        """
        {% macro script(this, kwargs) %}
        var {{ this.get_name() }} = L.Control.extend({
            onAdd: function(map) {
                var control;
                if (this.options.imagePath){
                    var img = document.createElement('img');
                    img.src=this.options.imagePath;
                    img.style.width = (this.options.scale * 100) + 'px';
                    img.style.height = (this.options.scale * 100) + 'px';
                    control=img;
                } else {
                    var template = document.createElement('template');
                    template.innerHTML = `{{ this.html }}`.trim();
                    control = template.content.firstChild
                }

                control.style.transform = 'rotate({{ this.angle }}deg)';
                return control;
            }
        });
        (new {{ this.get_name() }}({{ this.options|tojson }})).addTo({{this._parent.get_name()}});

        {% endmacro %}
    """
    )

    def __init__(self, html, imagePath="", position="bottomright", angle="", scale=""):
        super().__init__(html=html, position=position)
        self.angle = angle
        self.options = folium.utilities.parse_options(imagePath=imagePath, scale=scale, position=position)


class ScaleElement(MacroElement):
    """
    Class to wrap arbitrary HTML as Leaflet Control.

    Parameters
    ----------
    html : str
        HTML to render an element from.
    position : str
        Possible values are 'topleft', 'topright', 'bottomleft' or 'bottomright'.

    """

    _template = Template(
        """
        {% macro script(this, kwargs) %}

        var {{ this.get_name() }} = L.Control.Scale.extend({
            onAdd: function(map) {

                var container = document.createElement('div');
                container.classList.add("leaflet-control-scale");
                container.innerHTML = `{{ this.html }}`.trim();

                this._scale = container.firstChild;

                map.on('move', this._update, this);
                map.whenReady(this._update, this);

                return container;
            },

            _updateImperial(maxMeters) {
                const maxFeet = maxMeters * 3.2808399;
                let maxMiles, miles, feet;

                if (maxFeet > 5280) {
                    maxMiles = maxFeet / 5280;
                    miles = this._getRoundNum(maxMiles);
                    this._updateScale(this._scale, miles, "mi", miles / maxMiles);

                } else {
                    feet = this._getRoundNum(maxFeet);
                    this._updateScale(this._scale, feet, "ft", feet / maxFeet);
                }
            },

            _updateMetric(maxMeters) {
                const meters = this._getRoundNum(maxMeters),
                label = meters < 1000 ? `${meters} m` : `${meters / 1000} km`;

                value = meters < 1000 ? meters : meters / 1000;
                unit = meters < 1000 ? `m` : `km`;

                this._updateScale(this._scale, value, unit, meters / maxMeters);
            },

            _updateScale(scale, value, unit, ratio) {
                scale.style.width = `${Math.round(this.options.maxWidth * ratio * (4/3))}px`;

                scale.getElementById("scale").textContent = value;
                scale.getElementById("half_scale").textContent = value / 2;
                scale.getElementById("unit").textContent = unit;

            },

            onRemove(map) {
                map.off('move', this._update, this);
            }

        });
        (new {{ this.get_name() }}({{ this.options|tojson }})).addTo({{this._parent.get_name()}});

        {% endmacro %}
    """
    )

    def __init__(self, html, position="bottomright", imperial=False):
        super().__init__()
        self.html = html
        self.options = folium.utilities.parse_options(
            position=position, maxWidth=300, imperial=imperial, metric=not imperial
        )


class FloatElement(MacroElement):
    """
    Class to wrap arbitrary HTML as a floating Element.

    Parameters
    ----------
    html : str
        HTML to render an element from.
    left, right, top, bottom : str
        Distance between edge of map and nearest edge of element. Two should be provided. Style can also be specified
        in html.

    """

    _template = Template(
        """
        {% macro header(this,kwargs) %}
        <style>
            #{{ this.get_name() }} {
                position: absolute;
                z-index: 9999;
                left: {{ this.left }};
                right: {{ this.right }};
                top: {{ this.top }};
                bottom: {{ this.bottom }};
            }
        </style>
        {% endmacro %}

        {% macro html(this,kwargs) %}
        <div id="{{ this.get_name() }}">
             {{ this.html }}
        </div>
        {% endmacro %}
    """
    )

    def __init__(self, html, left="", right="", top="", bottom=""):
        super().__init__()
        self.html = html
        self.left = left
        self.right = right
        self.top = top
        self.bottom = bottom


class GeoTIFFElement(MacroElement):
    """
    Class to display a GeoTIFF.

    Parameters
    ----------
    url : str
        URL of GeoTIFF
    zoom : bool
        Zoom to displayed GeoTIFF
    """

    _template = Template(
        """
    {% macro html(this, kwargs) %}
        <script src="https://unpkg.com/georaster-layer-for-leaflet@3.8.0/dist/georaster-layer-for-leaflet.min.js"></script>
        <script src="https://unpkg.com/georaster@1.5.6/dist/georaster.browser.bundle.min.js"></script>
    {% endmacro %}

    {% macro script(this, kwargs) %}
        fetch(`{{ this.url }}`)
        .then(response => response.arrayBuffer())
        .then(arrayBuffer => {
          parseGeoraster(arrayBuffer).then(georaster => {
            var layer = new GeoRasterLayer({
                georaster: georaster,
                opacity: 0.7
            });
            layer.addTo({{this._parent.get_name()}});

            if ("True" == `{{ this.zoom }}`) {
                ({{ this._parent.get_name() }}).fitBounds(layer.getBounds());
            }

        });
        });
    {% endmacro %}
    """  # noqa
    )

    def __init__(self, url, zoom=False):
        super().__init__()
        self.url = url
        self.zoom = zoom


class PrintControl(MacroElement):
    _template = Template(
        """
                {% macro header(this, kwargs) %}
                <link rel="stylesheet" href="https://unpkg.com/leaflet.bigimage@1.0.1/src/Leaflet.BigImage.css">
                <script src='https://unpkg.com/leaflet.bigimage@1.0.1/src/Leaflet.BigImage.js'></script>
                {% endmacro %}
                {% macro script(this, kwargs) %}
                    L.control.BigImage().addTo(({{ this._parent.get_name() }}));
                {% endmacro %}
        """  # noqa
    )



import pandas as pd
import pytest

from ecoscope.base.utils import ModisBegin


@pytest.mark.skip(reason="this has been failing since May 2022; will be fixed in a follow-up pull")
def test_modis_offset():
    ts1 = pd.Timestamp("2022-01-13 17:00:00+0")
    ts2 = pd.Timestamp("2022-12-26 17:00:00+0")
    assert ts1 + ModisBegin() == pd.Timestamp("2022-01-17 00:00:00+0")
    assert ts2 + ModisBegin() == pd.Timestamp("2023-01-01 00:00:00+0")



"""
MIT License

Copyright (c) 2020-2021, Qiusheng Wu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import ee
import os
import json

__version__ = "0.30.0"


def in_colab_shell():
    """Tests if the code is being executed within Google Colab."""
    import sys

    if "google.colab" in sys.modules:
        return True
    else:
        return False


def ee_initialize(
    token_name="EARTHENGINE_TOKEN",
    auth_mode="notebook",
    service_account=False,
    auth_args={},
    user_agent_prefix="geemap",
    **kwargs,
):
    """Authenticates Earth Engine and initialize an Earth Engine session

    Args:
        token_name (str, optional): The name of the Earth Engine token. Defaults to "EARTHENGINE_TOKEN".
        auth_mode (str, optional): The authentication mode, can be one of paste,notebook,gcloud,appdefault. Defaults to "notebook".
        service_account (bool, optional): If True, use a service account. Defaults to False.
        auth_args (dict, optional): Additional authentication parameters for aa.Authenticate(). Defaults to {}.
        user_agent_prefix (str, optional): If set, the prefix (version-less) value used for setting the user-agent string. Defaults to "geemap".
        kwargs (dict, optional): Additional parameters for ee.Initialize(). For example,
            opt_url='https://earthengine-highvolume.googleapis.com' to use the Earth Engine High-Volume platform. Defaults to {}.
    """
    import httplib2

    user_agent = f"{user_agent_prefix}/{__version__}"
    if "http_transport" not in kwargs:
        kwargs["http_transport"] = httplib2.Http()

    auth_args["auth_mode"] = auth_mode

    if ee.data._credentials is None:
        ee_token = os.environ.get(token_name)
        if in_colab_shell():
            from google.colab import userdata

            try:
                ee_token = userdata.get(token_name)
            except Exception:
                pass
        if service_account:
            try:
                credential_file_path = os.path.expanduser(
                    "~/.config/earthengine/private-key.json"
                )

                if os.path.exists(credential_file_path):
                    with open(credential_file_path) as f:
                        token_dict = json.load(f)
                else:
                    token_name = "EARTHENGINE_TOKEN"
                    ee_token = os.environ.get(token_name)
                    token_dict = json.loads(ee_token, strict=False)
                service_account = token_dict["client_email"]
                private_key = token_dict["private_key"]

                credentials = ee.ServiceAccountCredentials(
                    service_account, key_data=private_key
                )
                ee.Initialize(credentials, **kwargs)

            except Exception as e:
                raise Exception(e)

        else:
            try:
                if ee_token is not None:
                    credential_file_path = os.path.expanduser(
                        "~/.config/earthengine/credentials"
                    )
                    if not os.path.exists(credential_file_path):
                        os.makedirs(
                            os.path.dirname(credential_file_path), exist_ok=True
                        )
                        if ee_token.startswith("{") and ee_token.endswith(
                            "}"
                        ):  # deals with token generated by new auth method (earthengine-api>=0.1.304).
                            token_dict = json.loads(ee_token)
                            with open(credential_file_path, "w") as f:
                                f.write(json.dumps(token_dict))
                        else:
                            credential = (
                                '{"refresh_token":"%s"}' % ee_token
                            )  # deals with token generated by old auth method.
                            with open(credential_file_path, "w") as f:
                                f.write(credential)

                ee.Initialize(**kwargs)

            except Exception:
                ee.Authenticate(**auth_args)
                ee.Initialize(**kwargs)

    ee.data.setUserAgent(user_agent)



from ecoscope.mapping.map import (
    ControlElement,
    EcoMap,
    FloatElement,
    NorthArrowElement,
    ScaleElement,
    GeoTIFFElement,
    PrintControl,
)

__all__ = [
    "ControlElement",
    "EcoMap",
    "FloatElement",
    "NorthArrowElement",
    "ScaleElement",
    "GeoTIFFElement",
    "PrintControl",
]



import os

import geopandas as gpd
import geopandas.testing
import numpy as np
import pandas as pd
import pytest
import sklearn.preprocessing

import ecoscope
from ecoscope.analysis.ecograph import (
    FeatureNameError,
    IndividualNameError,
    InterpolationError,
)


@pytest.mark.skip(reason="this has been failing since May 2022; will be fixed in a follow-up pull")
def test_ecograph(movbank_relocations):
    # apply relocation coordinate filter to movbank data
    pnts_filter = ecoscope.base.RelocsCoordinateFilter(
        min_x=-5,
        max_x=1,
        min_y=12,
        max_y=18,
        filter_point_coords=[[180, 90], [0, 0]],
    )
    movbank_relocations.apply_reloc_filter(pnts_filter, inplace=True)
    movbank_relocations.remove_filtered(inplace=True)

    # Create Trajectory
    movebank_trajectory_gdf = ecoscope.base.Trajectory.from_relocations(movbank_relocations)

    os.makedirs("tests/outputs", exist_ok=True)

    mean_step_length = np.mean(np.abs(movebank_trajectory_gdf["dist_meters"]))
    ecograph = ecoscope.analysis.Ecograph(movebank_trajectory_gdf, resolution=mean_step_length)

    ecograph.to_geotiff(
        "betweenness",
        "tests/outputs/salif_btwn_max.tif",
        individual="Salif Keita",
        interpolation="max",
    )
    salif_btwn_max_gdf = ecoscope.analysis.ecograph.get_feature_gdf("tests/outputs/salif_btwn_max.tif")

    ecograph.to_geotiff(
        "degree",
        "tests/outputs/salif_degree_mean.tif",
        individual="Salif Keita",
        interpolation="mean",
    )
    salif_degree_mean_gdf = ecoscope.analysis.ecograph.get_feature_gdf("tests/outputs/salif_degree_mean.tif")

    ecograph.to_geotiff(
        "collective_influence",
        "tests/outputs/salif_ci_min.tif",
        individual="Salif Keita",
        interpolation="min",
    )
    salif_ci_min_gdf = ecoscope.analysis.ecograph.get_feature_gdf("tests/outputs/salif_ci_min.tif")

    ecograph.to_geotiff(
        "step_length",
        "tests/outputs/salif_sl_median.tif",
        individual="Salif Keita",
        interpolation="median",
    )
    salif_sl_median_gdf = ecoscope.analysis.ecograph.get_feature_gdf("tests/outputs/salif_sl_median.tif")

    ecograph.to_geotiff(
        "degree",
        "tests/outputs/salif_degree_mean_std.tif",
        individual="Salif Keita",
        interpolation="mean",
        transform=sklearn.preprocessing.StandardScaler(),
    )
    salif_degree_mean_std_gdf = ecoscope.analysis.ecograph.get_feature_gdf("tests/outputs/salif_degree_mean_std.tif")

    ecograph.to_geotiff("speed", "tests/outputs/salif_speed.tif", individual="Salif Keita")
    salif_speed_gdf = ecoscope.analysis.ecograph.get_feature_gdf("tests/outputs/salif_speed.tif")

    ecograph.to_geotiff("dot_product", "tests/outputs/salif_dotprod.tif", individual="Salif Keita")
    salif_dotprod_gdf = ecoscope.analysis.ecograph.get_feature_gdf("tests/outputs/salif_dotprod.tif")

    expected_btw_max = gpd.read_feather("tests/test_output/salif_btwn_max.feather")
    expected_degree_mean = gpd.read_feather("tests/test_output/salif_degree_mean.feather")
    expected_ci_min = gpd.read_feather("tests/test_output/salif_ci_min.feather")
    expected_sl_median = gpd.read_feather("tests/test_output/salif_sl_median.feather")
    expected_speed = gpd.read_feather("tests/test_output/salif_speed.feather")
    expected_dotprod = gpd.read_feather("tests/test_output/salif_dotprod.feather")
    expected_degree_mean_std = gpd.read_feather("tests/test_output/salif_degree_mean_std.feather")

    gpd.testing.assert_geodataframe_equal(salif_btwn_max_gdf, expected_btw_max, check_less_precise=True)
    gpd.testing.assert_geodataframe_equal(salif_degree_mean_gdf, expected_degree_mean, check_less_precise=True)
    gpd.testing.assert_geodataframe_equal(salif_ci_min_gdf, expected_ci_min, check_less_precise=True)
    gpd.testing.assert_geodataframe_equal(salif_sl_median_gdf, expected_sl_median, check_less_precise=True)
    gpd.testing.assert_geodataframe_equal(salif_speed_gdf, expected_speed, check_less_precise=True)
    gpd.testing.assert_geodataframe_equal(salif_dotprod_gdf, expected_dotprod, check_less_precise=True)
    gpd.testing.assert_geodataframe_equal(salif_degree_mean_std_gdf, expected_degree_mean_std, check_less_precise=True)

    with pytest.raises(InterpolationError):
        ecograph.to_geotiff(
            "betweenness",
            "tests/outputs/salif_btwn_max.tif",
            individual="Salif Keita",
            interpolation="osjfos",
        )
    with pytest.raises(FeatureNameError):
        ecograph.to_geotiff(
            "rojfoofj",
            "tests/outputs/salif_btwn_max.tif",
            individual="Salif Keita",
            interpolation="max",
        )
    with pytest.raises(IndividualNameError):
        ecograph.to_geotiff(
            "betweenness",
            "tests/outputs/salif_btwn_max.tif",
            individual="fofkfp",
            interpolation="max",
        )

    ecograph.to_csv("tests/outputs/features.csv")
    feat = pd.read_csv("tests/outputs/features.csv")
    expected_feat = pd.read_csv("tests/test_output/features.csv")
    pd.testing.assert_frame_equal(feat, expected_feat)



"""
MIT License

Copyright (c) 2021, Qiusheng Wu

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__version__ = "be6ccdfa5450b775724547b6ef50f213c820be85"

"""Module for basemaps. 

More WMS basemaps can be found at the following websites:

1. USGS National Map: https://viewer.nationalmap.gov/services

2. MRLC NLCD Land Cover data: https://www.mrlc.gov/data-services-page

3. FWS NWI Wetlands data: https://www.fws.gov/wetlands/Data/Web-Map-Services.html

"""
import collections
import folium
import xyzservices.providers as xyz

# from box import Box

# Custom XYZ tile services.
xyz_tiles = {
    "OpenStreetMap": {
        "url": "https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png",
        "attribution": "OpenStreetMap",
        "name": "OpenStreetMap",
    },
    "ROADMAP": {
        "url": "https://mt1.google.com/vt/lyrs=m&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Maps",
    },
    "SATELLITE": {
        "url": "https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Satellite",
    },
    "TERRAIN": {
        "url": "https://mt1.google.com/vt/lyrs=p&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Terrain",
    },
    "HYBRID": {
        "url": "https://mt1.google.com/vt/lyrs=y&x={x}&y={y}&z={z}",
        "attribution": "Google",
        "name": "Google Satellite",
    },
}

# Custom WMS tile services.
wms_tiles = {
    "FWS NWI Wetlands": {
        "url": "https://www.fws.gov/wetlands/arcgis/services/Wetlands/MapServer/WMSServer?",
        "layers": "1",
        "name": "FWS NWI Wetlands",
        "attribution": "FWS",
        "format": "image/png",
        "transparent": True,
    },
    "FWS NWI Wetlands Raster": {
        "url": "https://www.fws.gov/wetlands/arcgis/services/Wetlands_Raster/ImageServer/WMSServer?",
        "layers": "0",
        "name": "FWS NWI Wetlands Raster",
        "attribution": "FWS",
        "format": "image/png",
        "transparent": True,
    },
    "NLCD 2019 CONUS Land Cover": {
        "url": "https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2019_Land_Cover_L48/wms?",
        "layers": "NLCD_2019_Land_Cover_L48",
        "name": "NLCD 2019 CONUS Land Cover",
        "attribution": "MRLC",
        "format": "image/png",
        "transparent": True,
    },
    "NLCD 2016 CONUS Land Cover": {
        "url": "https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2016_Land_Cover_L48/wms?",
        "layers": "NLCD_2016_Land_Cover_L48",
        "name": "NLCD 2016 CONUS Land Cover",
        "attribution": "MRLC",
        "format": "image/png",
        "transparent": True,
    },
    "NLCD 2013 CONUS Land Cover": {
        "url": "https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2013_Land_Cover_L48/wms?",
        "layers": "NLCD_2013_Land_Cover_L48",
        "name": "NLCD 2013 CONUS Land Cover",
        "attribution": "MRLC",
        "format": "image/png",
        "transparent": True,
    },
    "NLCD 2011 CONUS Land Cover": {
        "url": "https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2011_Land_Cover_L48/wms?",
        "layers": "NLCD_2011_Land_Cover_L48",
        "name": "NLCD 2011 CONUS Land Cover",
        "attribution": "MRLC",
        "format": "image/png",
        "transparent": True,
    },
    "NLCD 2008 CONUS Land Cover": {
        "url": "https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2008_Land_Cover_L48/wms?",
        "layers": "NLCD_2008_Land_Cover_L48",
        "name": "NLCD 2008 CONUS Land Cover",
        "attribution": "MRLC",
        "format": "image/png",
        "transparent": True,
    },
    "NLCD 2006 CONUS Land Cover": {
        "url": "https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2006_Land_Cover_L48/wms?",
        "layers": "NLCD_2006_Land_Cover_L48",
        "name": "NLCD 2006 CONUS Land Cover",
        "attribution": "MRLC",
        "format": "image/png",
        "transparent": True,
    },
    "NLCD 2004 CONUS Land Cover": {
        "url": "https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2004_Land_Cover_L48/wms?",
        "layers": "NLCD_2004_Land_Cover_L48",
        "name": "NLCD 2004 CONUS Land Cover",
        "attribution": "MRLC",
        "format": "image/png",
        "transparent": True,
    },
    "NLCD 2001 CONUS Land Cover": {
        "url": "https://www.mrlc.gov/geoserver/mrlc_display/NLCD_2001_Land_Cover_L48/wms?",
        "layers": "NLCD_2001_Land_Cover_L48",
        "name": "NLCD 2001 CONUS Land Cover",
        "attribution": "MRLC",
        "format": "image/png",
        "transparent": True,
    },
    "USGS NAIP Imagery": {
        "url": "https://imagery.nationalmap.gov/arcgis/services/USGSNAIPImagery/ImageServer/WMSServer?",
        "layers": "USGSNAIPImagery:NaturalColor",
        "name": "USGS NAIP Imagery",
        "attribution": "USGS",
        "format": "image/png",
        "transparent": True,
    },
    "USGS NAIP Imagery False Color": {
        "url": "https://imagery.nationalmap.gov/arcgis/services/USGSNAIPImagery/ImageServer/WMSServer?",
        "layers": "USGSNAIPImagery:FalseColorComposite",
        "name": "USGS NAIP Imagery False Color",
        "attribution": "USGS",
        "format": "image/png",
        "transparent": True,
    },
    "USGS NAIP Imagery NDVI": {
        "url": "https://imagery.nationalmap.gov/arcgis/services/USGSNAIPImagery/ImageServer/WMSServer?",
        "layers": "USGSNAIPImagery:NDVI_Color",
        "name": "USGS NAIP Imagery NDVI",
        "attribution": "USGS",
        "format": "image/png",
        "transparent": True,
    },
    "USGS Hydrography": {
        "url": "https://basemap.nationalmap.gov/arcgis/services/USGSHydroCached/MapServer/WMSServer?",
        "layers": "0",
        "name": "USGS Hydrography",
        "attribution": "USGS",
        "format": "image/png",
        "transparent": True,
    },
    "USGS 3DEP Elevation": {
        "url": "https://elevation.nationalmap.gov/arcgis/services/3DEPElevation/ImageServer/WMSServer?",
        "layers": "33DEPElevation:Hillshade Elevation Tinted",
        "name": "USGS 3DEP Elevation",
        "attribution": "USGS",
        "format": "image/png",
        "transparent": True,
    },
    "ESA WorldCover 2020": {
        "url": "https://services.terrascope.be/wms/v2",
        "layers": "WORLDCOVER_2020_MAP",
        "name": "ESA Worldcover 2020",
        "attribution": "ESA",
        "format": "image/png",
        "transparent": True,
    },
    "ESA WorldCover 2020 S2 FCC": {
        "url": "https://services.terrascope.be/wms/v2",
        "layers": "WORLDCOVER_2020_S2_FCC",
        "name": "ESA Worldcover 2020 S2 FCC",
        "attribution": "ESA",
        "format": "image/png",
        "transparent": True,
    },
    "ESA WorldCover 2020 S2 TCC": {
        "url": "https://services.terrascope.be/wms/v2",
        "layers": "WORLDCOVER_2020_S2_TCC",
        "name": "ESA Worldcover 2020 S2 TCC",
        "attribution": "ESA",
        "format": "image/png",
        "transparent": True,
    },
}


def _unpack_sub_parameters(var, param):
    temp = var
    for sub_param in param.split("."):
        temp = getattr(temp, sub_param)
    return temp


def get_xyz_dict(free_only=True):
    """Returns a dictionary of xyz services.

    Args:
        free_only (bool, optional): Whether to return only free xyz tile services that do not require an access token. Defaults to True.

    Returns:
        dict: A dictionary of xyz services.
    """

    xyz_dict = {}
    for item in xyz.values():
        try:
            name = item["name"]
            tile = _unpack_sub_parameters(xyz, name)
            if _unpack_sub_parameters(xyz, name).requires_token():
                if free_only:
                    pass
                else:
                    xyz_dict[name] = tile
            else:
                xyz_dict[name] = tile

        except Exception:
            for sub_item in item:
                name = item[sub_item]["name"]
                tile = _unpack_sub_parameters(xyz, name)
                if _unpack_sub_parameters(xyz, name).requires_token():
                    if free_only:
                        pass
                    else:
                        xyz_dict[name] = tile
                else:
                    xyz_dict[name] = tile

    xyz_dict = collections.OrderedDict(sorted(xyz_dict.items()))
    return xyz_dict


def xyz_to_folium():
    """Convert xyz tile services to folium tile layers.

    Returns:
        dict: A dictionary of folium tile layers.
    """
    folium_dict = {}

    for key in xyz_tiles:
        name = xyz_tiles[key]["name"]
        url = xyz_tiles[key]["url"]
        attribution = xyz_tiles[key]["attribution"]
        folium_dict[key] = folium.TileLayer(
            tiles=url,
            attr=attribution,
            name=name,
            overlay=True,
            control=True,
            max_zoom=22,
        )

    for key in wms_tiles:
        name = wms_tiles[key]["name"]
        url = wms_tiles[key]["url"]
        layers = wms_tiles[key]["layers"]
        fmt = wms_tiles[key]["format"]
        transparent = wms_tiles[key]["transparent"]
        attribution = wms_tiles[key]["attribution"]
        folium_dict[key] = folium.WmsTileLayer(
            url=url,
            layers=layers,
            name=name,
            attr=attribution,
            fmt=fmt,
            transparent=transparent,
            overlay=True,
            control=True,
        )

    xyz_dict = get_xyz_dict()
    for item in xyz_dict:
        name = xyz_dict[item].name
        url = xyz_dict[item].build_url()
        attribution = xyz_dict[item].attribution
        if "max_zoom" in xyz_dict[item].keys():
            max_zoom = xyz_dict[item]["max_zoom"]
        else:
            max_zoom = 22
        folium_dict[name] = folium.TileLayer(
            tiles=url,
            attr=attribution,
            name=name,
            max_zoom=max_zoom,
            overlay=True,
            control=True,
        )

    return folium_dict



import geopandas as gpd
import pytest

import ecoscope

if not pytest.earthengine:
    pytest.skip(
        "Skipping tests because connection to Earth Engine is not available.",
        allow_module_level=True,
    )


@pytest.mark.skip(reason="this has been failing since May 2022; will be fixed in a follow-up pull")
def test_seasons():
    gdf = gpd.read_file("tests/sample_data/vector/AOI_sites.gpkg").to_crs(4326)

    aoi = gdf.geometry.iat[0]

    # Extract the standardized NDVI ndvi_vals within the AOI
    ndvi_vals = ecoscope.analysis.seasons.std_ndvi_vals(aoi, start="2010-01-01", end="2021-01-01")

    # Calculate the seasonal transition point
    cuts = ecoscope.analysis.seasons.val_cuts(ndvi_vals, 2)

    # Determine the seasonal time windows
    windows = ecoscope.analysis.seasons.seasonal_windows(ndvi_vals, cuts, season_labels=["dry", "wet"])

    ecoscope.plotting.plot_seasonal_dist(ndvi_vals["NDVI"], cuts)

    assert len(windows) > 0



import itertools
import typing
import uuid
from dataclasses import dataclass, field

import geopandas as gpd
import pandas as pd
from shapely.geometry import Point


@dataclass
class SpatialFeature:
    """
    A spatial geometry with an associated name and unique ID. Becomes a useful construct in several movdata calculations
    """

    name: str = ""
    unique_id: typing.Any = uuid.uuid4()
    geometry: typing.Any = None


@dataclass
class ProximityProfile:
    spatial_features: typing.List[SpatialFeature] = field(default=list)


class Proximity:
    @classmethod
    def calculate_proximity(cls, proximity_profile, trajectory):
        """
        A function to analyze the trajectory of a subject in relation to a set of spatial features and regions to
        determine where/when the subject was proximal to the spatial feature.

        Parameters
        ----------
        proximity_profile: ProximityProfile
            proximity setting for performing calculation
        trajectory: ecoscope.base.Trajectory
            Geodataframe stores goemetry, speed_kmhr, heading etc. for each subject.
        Returns
        -------
        pd.DataFrame

        """
        proximity_events = []

        def analysis(traj):
            for sf in proximity_profile.spatial_features:
                proximity_dist = traj.geometry.distance(sf.geometry)
                start_fix = gpd.GeoSeries([Point(g.coords[0]) for g in traj.geometry])

                pr = traj[["groupby_col", "speed_kmhr", "heading"]]
                pr["proximity_distance"] = proximity_dist
                pr["proximal_fix"] = start_fix  # TODO: figure out the estimated fix interpolated along the seg
                pr["estimated_time"] = traj.segment_start
                pr["geometry"] = traj.geometry
                pr["spatialfeature_id"] = list(itertools.repeat(sf.unique_id, pr.shape[0]))
                pr["spatialfeature_name"] = list(itertools.repeat(sf.name, pr.shape[0]))

                proximity_events.append(pr)

        trajectory.groupby("groupby_col").apply(analysis)
        return pd.concat(proximity_events).reset_index(drop=True)



import datetime
import uuid

import geopandas as gpd
import pandas as pd
import pytest
from shapely.geometry import Point

import ecoscope

if not pytest.earthranger:
    pytest.skip(
        "Skipping tests because connection to EarthRanger is not available.",
        allow_module_level=True,
    )


def test_get_subject_observations(er_io):
    relocations = er_io.get_subject_observations(
        subject_ids=er_io.SUBJECT_IDS,
        include_subject_details=True,
        include_source_details=True,
        include_subjectsource_details=True,
    )
    assert not relocations.empty
    assert isinstance(relocations, ecoscope.base.Relocations)
    assert "groupby_col" in relocations
    assert "fixtime" in relocations
    assert "extra__source" in relocations


def test_get_source_observations(er_io):
    relocations = er_io.get_source_observations(
        source_ids=er_io.SOURCE_IDS,
        include_source_details=True,
    )
    assert isinstance(relocations, ecoscope.base.Relocations)
    assert "fixtime" in relocations
    assert "groupby_col" in relocations


def test_get_source_no_observations(er_io):
    relocations = er_io.get_source_observations(
        source_ids=str(uuid.uuid4()),
        include_source_details=True,
    )
    assert relocations.empty


def test_get_subjectsource_observations(er_io):
    relocations = er_io.get_subjectsource_observations(
        subjectsource_ids=er_io.SUBJECTSOURCE_IDS,
        include_source_details=True,
    )
    assert isinstance(relocations, ecoscope.base.Relocations)
    assert "fixtime" in relocations
    assert "groupby_col" in relocations


def test_get_subjectsource_no_observations(er_io):
    relocations = er_io.get_subjectsource_observations(
        subjectsource_ids=str(uuid.uuid4()),
        include_source_details=True,
    )
    assert relocations.empty


def test_get_subjectgroup_observations(er_io):
    relocations = er_io.get_subjectgroup_observations(group_name=er_io.GROUP_NAME)
    assert "groupby_col" in relocations


def test_get_events(er_events_io):
    events = er_events_io.get_events(event_type=["e00ce1f6-f9f1-48af-93c9-fb89ec493b8a"])
    assert not events.empty


def test_das_client_method(er_io):
    er_io.pulse()
    er_io.get_me()


def test_get_patrols(er_io):
    patrols = er_io.get_patrols()
    assert len(patrols) > 0


def test_post_observations(er_io):
    observations = gpd.GeoDataFrame.from_dict(
        [
            {
                "recorded_at": pd.Timestamp.utcnow().isoformat(),
                "geometry": Point(0, 0),
                "source": er_io.SOURCE_IDS[0],
            },
            {
                "recorded_at": (pd.Timestamp.utcnow() + pd.Timedelta(seconds=1)).isoformat(),
                "geometry": Point(0, 0),
                "source": er_io.SOURCE_IDS[0],
            },
            {
                "recorded_at": pd.Timestamp.utcnow().isoformat(),
                "geometry": Point(1, 1),
                "source": er_io.SOURCE_IDS[1],
            },
        ]
    )

    response = er_io.post_observations(observations)
    assert len(response) == 3
    assert "location" in response
    assert "recorded_at" in response


def test_post_events(er_io):
    events = [
        {
            "id": str(uuid.uuid4()),
            "title": "Accident",
            "event_type": "accident_rep",
            "time": pd.Timestamp.utcnow(),
            "location": {"latitude": -2.9553841592697982, "longitude": 38.033294677734375},
            "priority": 200,
            "state": "new",
            "event_details": {"type_accident": "head-on collision", "number_people_involved": 3, "animals_involved": 1},
            "is_collection": False,
            "icon_id": "accident_rep",
        },
        {
            "id": str(uuid.uuid4()),
            "title": "Accident",
            "event_type": "accident_rep",
            "time": pd.Timestamp.utcnow(),
            "location": {"latitude": -3.0321834919139206, "longitude": 38.4906005859375},
            "priority": 300,
            "state": "active",
            "event_details": {
                "type_accident": "side-impact collision",
                "number_people_involved": 2,
                "animals_involved": 1,
            },
            "is_collection": False,
            "icon_id": "accident_rep",
        },
    ]
    results = er_io.post_event(events)
    results["time"] = pd.to_datetime(results["time"], utc=True)

    expected = pd.DataFrame(events)
    results = results[expected.columns]
    pd.testing.assert_frame_equal(results, expected)


def test_patch_event(er_io):
    event = [
        {
            "id": str(uuid.uuid4()),
            "title": "Arrest",
            "event_type": "arrest_rep",
            "time": pd.Timestamp.utcnow(),
            "location": {"latitude": -3.4017015747197306, "longitude": 38.11809539794921},
            "priority": 200,
            "state": "new",
            "event_details": {
                "arrestrep_dateofbirth": "1985-01-1T13:00:00.000Z",
                "arrestrep_nationality": "other",
                "arrestrep_timeofarrest": datetime.datetime.utcnow().isoformat(),
                "arrestrep_reaonforarrest": "firearm",
                "arrestrep_arrestingranger": "catherine's cellphone",
            },
            "is_collection": False,
            "icon_id": "arrest_rep",
        }
    ]
    er_io.post_event(event)
    event_id = event[0]["id"]

    updated_event = pd.DataFrame(
        [
            {
                "priority": 300,
                "state": "active",
                "location": {"latitude": -4.135503657998179, "longitude": 38.4576416015625},
            }
        ]
    )

    result = er_io.patch_event(event_id=event_id, events=updated_event)
    result = result[["priority", "state", "location"]]
    pd.testing.assert_frame_equal(result, updated_event)


def test_get_patrol_observations(er_io):
    patrols = er_io.get_patrols()
    observations = er_io.get_patrol_observations(
        patrols,
        include_source_details=False,
        include_subject_details=False,
        include_subjectsource_details=False,
    )
    assert not observations.empty


def test_users(er_io):
    users = pd.DataFrame(er_io.get_users())
    assert not users.empty


def test_get_spatial_feature(er_io):
    spatial_feature = er_io.get_spatial_feature(spatial_feature_id="8868718f-0154-45bf-a74d-a66706ef958f")
    assert not spatial_feature.empty


def test_get_spatial_features_group(er_io):
    spatial_features = er_io.get_spatial_features_group(
        spatial_features_group_id="15698426-7e0f-41df-9bc3-495d87e2e097"
    )
    assert not spatial_features.empty



import datetime
import pytz
import json
import typing

import geopandas as gpd
import pandas as pd
import requests
from erclient.client import ERClient, ERClientException, ERClientNotFound
from tqdm.auto import tqdm

import ecoscope
from ecoscope.io.utils import pack_columns, to_hex


def fatal_status_code(e):
    return 400 <= e.response.status_code < 500


class EarthRangerIO(ERClient):
    def __init__(self, sub_page_size=4000, tcp_limit=5, **kwargs):
        if "server" in kwargs:
            server = kwargs.pop("server")
            kwargs["service_root"] = f"{server}/api/v1.0"
            kwargs["token_url"] = f"{server}/oauth2/token"

        self.sub_page_size = sub_page_size
        self.tcp_limit = tcp_limit
        kwargs["client_id"] = kwargs.get("client_id", "das_web_client")
        super().__init__(**kwargs)
        try:
            self.login()
        except ERClientNotFound:
            raise ERClientNotFound("Failed login. Check Stack Trace for specific reason.")

    def _token_request(self, payload):
        response = requests.post(self.token_url, data=payload)
        if response.ok:
            self.auth = json.loads(response.text)
            expires_in = int(self.auth["expires_in"]) - 5 * 60
            self.auth_expires = pytz.utc.localize(datetime.datetime.utcnow()) + datetime.timedelta(seconds=expires_in)
            return True

        self.auth = None
        self.auth_expires = pytz.utc.localize(datetime.datetime.min)
        raise ERClientNotFound(json.loads(response.text)["error_description"])

    @staticmethod
    def _clean_kwargs(addl_kwargs={}, **kwargs):
        for k in addl_kwargs.keys():
            print(f"Warning: {k} is a non-standard parameter. Results may be unexpected.")
        return {k: v for k, v in {**addl_kwargs, **kwargs}.items() if v is not None}

    @staticmethod
    def _normalize_column(df, col):
        print(col)
        for k, v in pd.json_normalize(df.pop(col), sep="__").add_prefix(f"{col}__").iteritems():
            df[k] = v.values

    @staticmethod
    def _dataframe_to_dict(events):
        if isinstance(events, gpd.GeoDataFrame):
            events["location"] = pd.DataFrame({"longitude": events.geometry.x, "latitude": events.geometry.y}).to_dict(
                "records"
            )
            del events["geometry"]

        if isinstance(events, pd.DataFrame):
            events = events.to_dict("records")
        return events

    @staticmethod
    def _to_gdf(df):
        longitude, latitude = (0, 1) if isinstance(df["location"].iat[0], list) else ("longitude", "latitude")
        return gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df["location"].str[longitude], df["location"].str[latitude]),
            crs=4326,
        )

    """
    GET Functions
    """

    def get_sources(
        self,
        manufacturer_id=None,
        provider_key=None,
        provider=None,
        id=None,
        **addl_kwargs,
    ):
        """
        Parameters
        ----------
        manufacturer_id
        provider_key
        provider
        id
        Returns
        -------
        sources : pd.DataFrame
            DataFrame of queried sources
        """

        params = self._clean_kwargs(
            addl_kwargs,
            manufacturer_id=manufacturer_id,
            provider_key=provider_key,
            provider=provider,
            id=id,
        )
        df = pd.DataFrame(
            self.get_objects_multithreaded(
                object="sources/", threads=self.tcp_limit, page_size=self.sub_page_size, **params
            )
        )
        assert not df.empty
        return df

    def get_subjects(
        self,
        include_inactive=None,
        bbox=None,
        subject_group=None,
        name=None,
        updated_since=None,
        tracks=None,
        id=None,
        updated_until=None,
        group_name=None,
        **addl_kwargs,
    ):
        """
        Parameters
        ----------
        include_inactive: Include inactive subjects in list.
        bbox: Include subjects having track data within this bounding box defined by a 4-tuple of coordinates marking
            west, south, east, north.
        subject_group: Indicate a subject group for which Subjects should be listed.
        name : Find subjects with the given name
        updated_since: Return Subject that have been updated since the given timestamp.
        tracks: Indicate whether to render each subject's recent tracks.
        id: A comma-delimited list of Subject IDs.
        updated_until
        group_name
        Returns
        -------
        subjects : pd.DataFrame
        """

        params = self._clean_kwargs(
            addl_kwargs,
            include_inactive=include_inactive,
            bbox=bbox,
            subject_group=subject_group,
            name=name,
            updated_since=updated_since,
            tracks=tracks,
            id=id,
            updated_until=updated_until,
            group_name=group_name,
        )

        assert params.get("subject_group") is None or params.get("group_name") is None

        if params.get("group_name") is not None:
            try:
                params["subject_group"] = self._get(
                    "subjectgroups/",
                    params={
                        "group_name": params.pop("group_name"),
                        "include_inactive": True,
                        "include_hidden": True,
                        "flat": True,
                    },
                )[0]["id"]
            except IndexError:
                raise KeyError("`group_name` not found")

        df = pd.DataFrame(
            self.get_objects_multithreaded(
                object="subjects/", threads=self.tcp_limit, page_size=self.sub_page_size, **params
            )
        )
        assert not df.empty

        df["hex"] = df["additional"].str["rgb"].map(to_hex) if "additional" in df else "#ff0000"

        return df

    def get_subjectsources(self, subjects=None, sources=None, **addl_kwargs):
        """
        Parameters
        ----------
        subjects: A comma-delimited list of Subject IDs.
        sources: A comma-delimited list of Source IDs.
        Returns
        -------
        subjectsources : pd.DataFrame
        """
        params = self._clean_kwargs(addl_kwargs, sources=sources, subjects=subjects)
        return pd.DataFrame(
            self.get_objects_multithreaded(
                object="subjectsources/", threads=self.tcp_limit, page_size=self.sub_page_size, **params
            )
        )

    def _get_observations(
        self,
        source_ids=None,
        subject_ids=None,
        subjectsource_ids=None,
        tz="UTC",
        since=None,
        until=None,
        filter=None,
        include_details=None,
        created_after=None,
        **addl_kwargs,
    ):
        """
        Return observations matching queries. If `subject_id`, `source_id`, or `subjectsource_id` is specified, the
        index is set to the provided value.
        Parameters
        ----------
        subject_ids: filter to a single subject
        source_ids: filter to a single source
        subjectsource_ids: filter to a subjectsource_id, rather than source_id + time range
        since: get observations after this ISO8061 date, include timezone
        until:get observations up to this ISO8061 date, include timezone
        filter
            filter using exclusion_flags for an observation.
            filter=None returns everything
            filter=0  filters out everything but rows with exclusion flag 0 (i.e, passes back clean data)
            filter=1  filters out everything but rows with exclusion flag 1 (i.e, passes back manually filtered data)
            filter=2, filters out everything but rows with exclusion flag 2 (i.e., passes back automatically filtered
            data)
            filter=3, filters out everything but rows with exclusion flag 2 or 1 (i.e., passes back both manual and
            automatically filtered data)
        include_details: one of [true,false], default is false. This brings back the observation additional field
        created_after: get observations created (saved in EarthRanger) after this ISO8061 date, include timezone
        Returns
        -------
        observations : gpd.GeoDataFrame
        """
        assert (source_ids, subject_ids, subjectsource_ids).count(None) == 2

        params = self._clean_kwargs(
            addl_kwargs,
            since=since,
            until=until,
            filter=filter,
            include_details=include_details,
            created_after=created_after,
        )

        if source_ids:
            id_name, ids = "source_id", source_ids
        elif subject_ids:
            id_name, ids = "subject_id", subject_ids
        else:
            id_name, ids = "subjectsource_id", subjectsource_ids

        observations = []
        pbar = tqdm([ids] if isinstance(ids, str) else ids)

        for _id in pbar:
            params[id_name] = _id
            pbar.set_description(f"Downloading Observations for {id_name}={_id}")
            dataframe = pd.DataFrame(
                self.get_objects_multithreaded(
                    object="observations/", threads=self.tcp_limit, page_size=self.sub_page_size, **params
                )
            )
            dataframe[id_name] = _id
            observations.append(dataframe)

        observations = pd.concat(observations)
        if observations.empty:
            return gpd.GeoDataFrame()

        observations["created_at"] = pd.to_datetime(
            observations["created_at"],
            errors="coerce",
            utc=True,
        ).dt.tz_convert(tz)

        observations["recorded_at"] = pd.to_datetime(
            observations["recorded_at"],
            errors="coerce",
            utc=True,
        ).dt.tz_convert(tz)

        observations.sort_values("recorded_at", inplace=True)
        return EarthRangerIO._to_gdf(observations)

    def get_source_observations(self, source_ids, include_source_details=False, relocations=True, **kwargs):
        """
        Get observations for each listed source and create a `Relocations` object.
        Parameters
        ----------
        source_ids : str or list[str]
            List of source UUIDs
        include_source_details : bool, optional
            Whether to merge source info into dataframe
        kwargs
            Additional arguments to pass in the request to EarthRanger. See the docstring of `_get_observations` for
            info.
        Returns
        -------
        relocations : ecoscope.base.Relocations
            Observations in `Relocations` format
        """

        if isinstance(source_ids, str):
            source_ids = [source_ids]

        observations = self._get_observations(source_ids=source_ids, **kwargs)
        if observations.empty:
            return gpd.GeoDataFrame()

        if include_source_details:
            observations = observations.merge(
                pd.DataFrame(self.get_sources(id=",".join(observations["source"].unique()))).add_prefix("source__"),
                left_on="source",
                right_on="source__id",
            )

        if relocations:
            return ecoscope.base.Relocations.from_gdf(
                observations,
                groupby_col="source",
                uuid_col="id",
                time_col="recorded_at",
            )
        else:
            return observations

    def get_subject_observations(
        self,
        subject_ids,
        include_source_details=False,
        include_subject_details=False,
        include_subjectsource_details=False,
        relocations=True,
        **kwargs,
    ):
        """
        Get observations for each listed subject and create a `Relocations` object.
        Parameters
        ----------
        subject_ids : str or list[str]
            List of subject UUIDs
        include_source_details : bool, optional
            Whether to merge source info into dataframe
        include_subject_details : bool, optional
            Whether to merge subject info into dataframe
        include_subjectsource_details : bool, optional
            Whether to merge subjectsource info into dataframe
        kwargs
            Additional arguments to pass in the request to EarthRanger. See the docstring of `__get_observations` for
            info.
        Returns
        -------
        relocations : ecoscope.base.Relocations
            Observations in `Relocations` format
        """

        if isinstance(subject_ids, str):
            subject_ids = [subject_ids]

        observations = self._get_observations(subject_ids=subject_ids, **kwargs)

        if observations.empty:
            return gpd.GeoDataFrame()

        if include_source_details:
            observations = observations.merge(
                self.get_sources(id=",".join(observations["source"].unique())).add_prefix("source__"),
                left_on="source",
                right_on="source__id",
            )
        if include_subject_details:
            observations = observations.merge(
                self.get_subjects(id=",".join(subject_ids), include_inactive=True).add_prefix("subject__"),
                left_on="subject_id",
                right_on="subject__id",
            )

        if include_subjectsource_details:
            observations = observations.merge(
                self.get_subjectsources(subjects=",".join(observations["subject_id"].unique())).add_prefix(
                    "subjectsource__"
                ),
                left_on=["subject_id", "source"],
                right_on=["subjectsource__subject", "subjectsource__source"],
            )

        if relocations:
            return ecoscope.base.Relocations.from_gdf(
                observations,
                groupby_col="subject_id",
                uuid_col="id",
                time_col="recorded_at",
            )
        else:
            return observations

    def get_subjectsource_observations(
        self,
        subjectsource_ids,
        include_source_details=False,
        relocations=True,
        **kwargs,
    ):
        """
        Get observations for each listed subjectsource and create a `Relocations` object.
        Parameters
        ----------
        subjectsource_ids : str or list[str]
            List of subjectsource UUIDs
        include_source_details : bool, optional
            Whether to merge source info into dataframe
        kwargs
            Additional arguments to pass in the request to EarthRanger. See the docstring of `__get_observations` for
            info.
        Returns
        -------
        relocations : ecoscope.base.Relocations
            Observations in `Relocations` format
        """

        if isinstance(subjectsource_ids, str):
            subjectsource_ids = [subjectsource_ids]

        observations = self._get_observations(subjectsource_ids=subjectsource_ids, **kwargs)

        if observations.empty:
            return gpd.GeoDataFrame()

        if include_source_details:
            observations = observations.merge(
                pd.DataFrame(self.get_sources(id=",".join(observations["source"].unique()))).add_prefix("source__"),
                left_on="source",
                right_on="source__id",
            )

        if relocations:
            return ecoscope.base.Relocations.from_gdf(
                observations,
                groupby_col="subjectsource_id",
                uuid_col="id",
                time_col="recorded_at",
            )
        else:
            return observations

    def get_subjectgroup_observations(self, subject_group=None, group_name=None, include_inactive=True, **kwargs):
        """
        Parameters
        ----------
        subject_group : str
            UUID of subject group to filter by
        group_name : str
            Common name of subject group to filter by
        include_inactive : bool, optional
            Whether to get observations for Subjects marked inactive by EarthRanger
        kwargs
            Additional arguments to pass in the request to `get_subject_observations`. See the docstring of
            `get_subject_observations` for info.
        Returns
        -------
        relocations : ecoscope.base.Relocations
            Observations in `Relocations` format
        """

        assert (subject_group is None) != (group_name is None)

        if subject_group:
            subject_ids = self.get_subjects(subject_group=subject_group, include_inactive=include_inactive).id.tolist()
        else:
            subject_ids = self.get_subjects(group_name=group_name, include_inactive=include_inactive).id.tolist()

        return self.get_subject_observations(subject_ids, **kwargs)

    def get_event_types(self, include_inactive=False, **addl_kwargs):
        params = self._clean_kwargs(addl_kwargs, include_inactive=include_inactive)

        return pd.DataFrame(self._get("activity/events/eventtypes", **params))

    def get_events(
        self,
        is_collection=None,
        updated_size=None,
        event_ids=None,
        bbox=None,
        sort_by=None,
        patrol_segment=None,
        state=None,
        event_type=None,
        include_updates=False,
        include_details=False,
        include_notes=False,
        include_related_events=False,
        include_files=False,
        max_results=None,
        oldest_update_date=None,
        exclude_contained=None,
        updated_since=None,
        event_category=None,
        since=None,
        until=None,
        **addl_kwargs,
    ):
        """
        Parameters
        ----------
        is_collection
            true/false whether to filter on is_collection
        updated_since
            date-string to limit on updated_at
        event_ids : array[string]
            Event IDs, comma-separated
        bbox
            bounding box including four coordinate values, comma-separated. Ex. bbox=-122.4,48.4,-122.95,49.0
            (west, south, east, north).
        sort_by
            Sort by (use 'event_time', 'updated_at', 'created_at', 'serial_number') with optional minus ('-') prefix to
            reverse order.
        patrol_segment
            ID of patrol segment to filter on
        state
            Comma-separated list of 'scheduled'/'active'/'overdue'/'done'/'cancelled'
        event_type
            Comma-separated list of event type uuids
        include_updates
            Boolean value
        include_details
            Boolean value
        include_notes
            Boolean value
        include_related_events
            Boolean value
        include_files
            Boolean value
        max_results
        oldest_update_date
        exclude_contained
        event_category
        since
        until
        Returns
        -------
        events : gpd.GeoDataFrame
            GeoDataFrame of queried events
        """

        params = self._clean_kwargs(
            addl_kwargs,
            is_collection=is_collection,
            updated_size=updated_size,
            event_ids=event_ids,
            bbox=bbox,
            sort_by=sort_by,
            patrol_segment=patrol_segment,
            state=state,
            event_type=event_type,
            include_updates=include_updates,
            include_details=include_details,
            include_notes=include_notes,
            include_related_events=include_related_events,
            include_files=include_files,
            max_results=max_results,
            oldest_update_date=oldest_update_date,
            exclude_contained=exclude_contained,
            updated_since=updated_since,
            event_category=event_category,
        )

        filter = {"date_range": {}}
        if since is not None:
            filter["date_range"]["lower"] = since
            params["filter"] = json.dumps(filter)
        if until is not None:
            filter["date_range"]["upper"] = until
            params["filter"] = json.dumps(filter)

        df = pd.DataFrame(
            self.get_objects_multithreaded(
                object="activity/events/", threads=self.tcp_limit, page_size=self.sub_page_size, **params
            )
        )

        assert not df.empty
        df["time"] = pd.to_datetime(df["time"])

        df = gpd.GeoDataFrame(df)
        df.loc[~df["geojson"].isna(), "geometry"] = gpd.GeoDataFrame.from_features(
            df.loc[~df["geojson"].isna(), "geojson"]
        )["geometry"]
        df.set_crs(4326, inplace=True)

        df.sort_values("time", inplace=True)
        return df.set_index("id")

    def get_patrol_types(self):
        df = pd.DataFrame(self._get("activity/patrols/types"))
        return df.set_index("id")

    def get_patrols(self, since=None, until=None, patrol_type=None, status=None, **addl_kwargs):
        """
        Parameters
        ----------
        since:
            lower date range
        until:
            upper date range
        patrol_type:
            Comma-separated list of type of patrol UUID
        status
            Comma-separated list of 'scheduled'/'active'/'overdue'/'done'/'cancelled'
        Returns
        -------
        patrols : pd.DataFrame
            DataFrame of queried patrols
        """

        params = self._clean_kwargs(
            addl_kwargs,
            status=status,
            patrol_type=[patrol_type] if isinstance(patrol_type, str) else patrol_type,
            return_data=True,
        )

        filter = {"date_range": {}, "patrol_type": []}

        if since is not None:
            filter["date_range"]["lower"] = since
        if until is not None:
            filter["date_range"]["upper"] = until
        if patrol_type is not None:
            filter["patrol_type"] = params["patrol_type"]
        params["filter"] = json.dumps(filter)

        df = pd.DataFrame(
            self.get_objects_multithreaded(
                object="activity/patrols", threads=self.tcp_limit, page_size=self.sub_page_size, **params
            )
        )
        if "serial_number" in df.columns:
            df = df.sort_values(by="serial_number").reset_index(drop=True)
        return df

    def get_patrol_segments_from_patrol_id(self, patrol_id, **addl_kwargs):
        """
        Download patrols for a given `patrol id`.

        Parameters
        ----------
        patrol_id :
            Patrol UUID.
        kwargs
            Additional parameters to pass to `_get`.

        Returns
        -------
        dataframe : Dataframe of patrols.
        """

        params = self._clean_kwargs(addl_kwargs)

        object = f"activity/patrols/{patrol_id}/"
        df = self._get(object, **params)
        df["patrol_segments"][0].pop("updates")
        df.pop("updates")

        return pd.DataFrame(dict([(k, pd.Series(v)) for k, v in df.items()]))

    def get_patrol_segments(self):
        object = "activity/patrols/segments/"
        return pd.DataFrame(
            self.get_objects_multithreaded(object=object, threads=self.tcp_limit, page_size=self.sub_page_size)
        )

    def get_patrol_observations(self, patrols_df, include_patrol_details=False, **kwargs):
        """
        Download observations for provided `patrols_df`.

        Parameters
        ----------
        patrols_df : pd.DataFrame
           Data returned from a call to `get_patrols`.
        include_patrol_details : bool, optional
           Whether to merge patrol details into dataframe
        kwargs
           Additional parameters to pass to `get_subject_observations`.

        Returns
        -------
        relocations : ecoscope.base.Relocations
        """

        observations = []
        for _, patrol in patrols_df.iterrows():
            for patrol_segment in patrol["patrol_segments"]:
                subject_id = (patrol_segment.get("leader") or {}).get("id")
                patrol_start_time = (patrol_segment.get("time_range") or {}).get("start_time")
                patrol_end_time = (patrol_segment.get("time_range") or {}).get("end_time")

                df_pt = self.get_patrol_types()
                patrol_type = df_pt[df_pt["value"] == patrol_segment.get("patrol_type")].reset_index()["id"][0]

                if None in {subject_id, patrol_start_time}:
                    continue

                try:
                    observation = self.get_subject_observations(
                        subject_ids=[subject_id], since=patrol_start_time, until=patrol_end_time, **kwargs
                    )
                    if include_patrol_details:
                        observation["patrol_id"] = patrol["id"]
                        observation["patrol_serial_number"] = patrol["serial_number"]
                        observation["patrol_start_time"] = patrol_start_time
                        observation["patrol_end_time"] = patrol_end_time
                        observation["patrol_type"] = patrol_type
                        observation = (
                            observation.reset_index()
                            .merge(
                                pd.DataFrame(self.get_patrol_types()).add_prefix("patrol_type__"),
                                left_on="patrol_type",
                                right_on="id",
                            )
                            .drop(
                                columns=[
                                    "patrol_type__ordernum",
                                    "patrol_type__icon_id",
                                    "patrol_type__default_priority",
                                    "patrol_type__is_active",
                                ]
                            )
                        )
                    if len(observation) > 0:
                        observations.append(observation)
                except Exception as e:
                    print(
                        f"Getting observations for subject_id={subject_id} start_time={patrol_start_time}"
                        f"end_time={patrol_end_time} failed for: {e}"
                    )

        df = ecoscope.base.Relocations(pd.concat(observations))
        if include_patrol_details:
            return df.set_index("id")
        return df

    def get_patrol_segment_events(
        self,
        patrol_segment_id=None,
        include_details=False,
        include_files=False,
        include_related_events=False,
        include_notes=False,
        **addl_kwargs,
    ):
        params = self._clean_kwargs(
            addl_kwargs,
            patrol_segment_id=patrol_segment_id,
            include_details=include_details,
            include_files=include_files,
            include_related_events=include_related_events,
            include_notes=include_notes,
        )

        object = f"activity/patrols/segments/{patrol_segment_id}/events/"
        return pd.DataFrame(
            self.get_objects_multithreaded(
                object=object, threads=self.tcp_limit, page_size=self.sub_page_size, **params
            )
        )

    def get_spatial_features_group(self, spatial_features_group_id=None, **addl_kwargs):
        """
        Download spatial features in a spatial features group for a given  `spatial features group id`.

        Parameters
        ----------
        spatial_features_group_id :
            Spatial Features Group UUID.
        kwargs
            Additional parameters to pass to `_get`.

        Returns
        -------
        dataframe : GeoDataFrame of spatial features in a spatial features group.
        """
        params = self._clean_kwargs(addl_kwargs, spatial_features_group_id=spatial_features_group_id)

        object = f"spatialfeaturegroup/{spatial_features_group_id}/"
        spatial_features_group = self._get(object, **params)

        spatial_features = []
        for spatial_feature in spatial_features_group["features"]:
            spatial_features.append(spatial_feature["features"][0])

        return gpd.GeoDataFrame.from_features(spatial_features)

    def get_spatial_feature(self, spatial_feature_id=None, **addl_kwargs):
        """
        Download spatial feature for a given  `spatial feature id`.

        Parameters
        ----------
        spatial_feature_id :
            Spatial Feature UUID.
        kwargs
            Additional parameters to pass to `_get`.

        Returns
        -------
        dataframe : GeoDataFrame of spatial feature.
        """

        params = self._clean_kwargs(addl_kwargs, spatial_feature_id=spatial_feature_id)

        object = f"spatialfeature/{spatial_feature_id}/"
        spatial_feature = self._get(object, **params)
        return gpd.GeoDataFrame.from_features(spatial_feature["features"])

    """
    POST Functions
    """

    def post_source(
        self,
        source_type: str,
        manufacturer_id: str,
        model_name: str,
        provider: str = "default",
        additional: typing.Dict = {},
        **kwargs,
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        source_type
        manufacturer_id
        model_name
        provider
        additional

        Returns
        -------
        pd.DataFrame
        """

        payload = {
            "source_type": source_type,
            "manufacturer_id": manufacturer_id,
            "model_name": model_name,
            "additional": additional,
            "provider": provider,
        }

        if kwargs:
            payload.update(kwargs)

        response = self._post("sources", payload=payload)
        return pd.DataFrame([response])

    def post_subject(
        self,
        subject_name: str,
        subject_type: str,
        subject_subtype: str,
        is_active: bool = True,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        subject_name
        subject_subtype
        is_active

        Returns
        -------
        pd.DataFrame
        """

        payload = {
            "name": subject_name,
            "subject_subtype": subject_subtype,
            "is_active": is_active,
        }

        if kwargs:
            payload.update(kwargs)

        response = self._post("subjects", payload=payload)
        return pd.DataFrame([response])

    def post_subjectsource(
        self,
        subject_id: str,
        source_id: str,
        lower_bound_assigned_range: datetime.datetime,
        upper_bound_assigned_range: datetime.datetime,
        additional: typing.Dict = None,
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        subject_id
        source_id
        lower_bound_assigned_range
        upper_bound_assigned_range
        additional

        Returns
        -------
        pd.DataFrame
        """

        if additional is None:
            additional = {}
        payload = {
            "source": source_id,
            "assigned_range": {
                "lower": lower_bound_assigned_range,
                "upper": upper_bound_assigned_range,
            },
            "additional": additional,
        }

        urlpath = f"subject/{subject_id}/sources"
        response = self._post(urlpath, payload=payload)
        return pd.DataFrame([response])

    def post_observations(
        self,
        observations: gpd.GeoDataFrame,
        source_id_col: str = "source",
        recorded_at_col: str = "recorded_at",
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        observations : gpd.GeoDataFrame
            observation data to be uploaded
        source_id_col : str
            The source column in the observation dataframe
        recorded_at_col : str
            The observation recorded time column in the dataframe

        Returns
        -------
        None
        """

        def upload(obs):
            try:
                obs = obs.rename(columns={source_id_col: "source", recorded_at_col: "recorded_at"})
                obs["location"] = pd.DataFrame({"longitude": obs.geometry.x, "latitude": obs.geometry.y}).to_dict(
                    "records"
                )
                del obs["geometry"]
                obs = pack_columns(obs, columns=["source", "recorded_at", "location"])
                post_data = obs.to_dict("records")
                results = super(EarthRangerIO, self).post_observation(post_data)
            except ERClientException as exc:
                self.logger.error(exc)
            except requests.exceptions.RequestException as exc:
                self.logger.error(exc)
            else:
                return pd.DataFrame(results)

        return observations.groupby(source_id_col, group_keys=False).progress_apply(upload)

    def post_event(
        self,
        events: typing.Union[gpd.GeoDataFrame, pd.DataFrame, typing.Dict, typing.List[typing.Dict]],
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        events

        Returns
        -------
        pd.DataFrame:
            New events created in EarthRanger.
        """

        events = self._dataframe_to_dict(events)
        results = super().post_event(event=events)
        results = results if isinstance(results, list) else [results]
        return pd.DataFrame(results)

    def post_patrol(self, priority: int, **kwargs) -> pd.DataFrame:
        """
        Parameters
        ----------
        priority

        Returns
        -------
        pd.DataFrame
        """
        payload = {"priority": priority}

        if kwargs:
            payload.update(kwargs)

        response = self._post("activity/patrols", payload=payload)
        return pd.DataFrame([response])

    def post_patrol_segment(
        self,
        patrol_id: str,
        patrol_segment_id: str,
        patrol_type: str = None,
        tracked_subject_id: str = None,
        scheduled_start: str = None,
        scheduled_end: str = None,
        start_time: str = None,
        end_time: str = None,
        start_location: typing.Tuple[float, float] = None,
        end_location: typing.Tuple[float, float] = None,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        patrol_id
        patrol_segment_id
        patrol_type
        tracked_subject_id
        scheduled_start
        scheduled_end
        start_time
        end_time
        start_location
        end_location

        Returns
        -------
        pd.DataFrame
        """

        payload = {
            "patrol": patrol_id,
            "patrol_segment": patrol_segment_id,
            "scheduled_start": scheduled_start,
            "scheduled_end": scheduled_end,
            "time_range": {"start_time": start_time, "end_time": end_time},
        }

        if tracked_subject_id is not None:
            payload.update({"leader": {"content_type": "observations.subject", "id": tracked_subject_id}})
        else:
            payload.update({"leader": None})

        if start_location is not None:
            payload.update({"start_location": {"latitude": start_location[0], "longitude": start_location[1]}})
        else:
            payload.update({"start_location": None})

        if end_location is not None:
            payload.update({"end_location": {"latitude": end_location[0], "longitude": end_location[1]}})
        else:
            payload.update({"end_location": None})

        if patrol_type is not None:
            payload.update({"patrol_type": patrol_type})

        if kwargs:
            payload.update(kwargs)

        response = self._post("activity/patrols/segments/", payload=payload)
        return pd.DataFrame([response])

    def post_patrol_segment_event(
        self,
        patrol_segment_id: str,
        event_type: str,
        **addl_kwargs,
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        patrol_segment_id
        event_type

        Returns
        -------
        pd.DataFrame
        """

        payload = {
            "patrol_segment": patrol_segment_id,
            "event_type": event_type,
        }

        if addl_kwargs:
            payload.update(addl_kwargs)

        response = self._post(f"activity/patrols/segments/{patrol_segment_id}/events/", payload=payload)
        return pd.DataFrame([response])

    """
    PATCH Functions
    """

    def patch_event(
        self,
        event_id: str,
        events: typing.Union[gpd.GeoDataFrame, pd.DataFrame, typing.Dict, typing.List[typing.Dict]],
    ) -> pd.DataFrame:
        """
        Parameters
        ----------
        event_id
            UUID for the event that will be updated.
        events

        Returns
        -------
        pd.DataFrame:
            Updated events in EarthRanger.
        """

        events = self._dataframe_to_dict(events)
        if isinstance(events, list):
            results = [self._patch(f"activity/event/{event_id}", payload=event) for event in events]
        else:
            results = [self._patch(f"activity/event/{event_id}", payload=events)]
        return pd.DataFrame(results)

    """
    DELETE Functions
    """

    def delete_observation(self, observation_id: str):
        """
        Parameters
        ----------
        observation_id
        -------
        """

        self._delete("observation/" + observation_id + "/")



import math
import os
import typing
from dataclasses import dataclass

import numba as nb
import numpy as np
import scipy
import sklearn
from scipy.optimize import minimize
from scipy.stats import weibull_min

from ecoscope.base import Trajectory
from ecoscope.io import raster


@nb.cfunc("double(intc, CPointer(double))")
def __etd__(_, a):
    s = a[0]
    k = a[1]
    el = a[2]
    m = a[3]

    return np.exp(-((s / el) ** k)) * s ** (k - 2) / np.sqrt(s**2 - m**2)


_etd = scipy.LowLevelCallable(__etd__.ctypes)


class WeibullPDF:
    @staticmethod
    def fit(data, floc=0):
        # returns estimates of shape and scale parameters from data; keeping the location fixed
        shape, _, scale = weibull_min.fit(data, floc=floc)
        return shape, scale

    @staticmethod
    def pdf(data, shape, location=0, scale=1):
        # probability density function.
        return weibull_min.pdf(x=data, c=shape, loc=location, scale=scale)

    @staticmethod
    def cdf(data, shape, location=0, scale=1):
        # cumulative distribution function.
        return weibull_min.cdf(x=data, c=shape, loc=location, scale=scale)

    @staticmethod
    def nelder_mead(func, x0, args=(), **kwargs):
        # minimization of scalar function of one or more variables using the Nelder-Mead algorithm.
        return minimize(fun=func, x0=x0, args=args, method="Nelder-Mead", **kwargs)

    @staticmethod
    def expected_func(speed, shape, scale, time, distance):
        # time-density expectation function for two-parameter weibull distribution.
        _funcs = [
            4 * shape / (math.pi * scale * speed),
            math.pow((speed / scale), shape - 1),
            math.exp(-1 * math.pow(speed / scale, shape))
            / math.sqrt(np.square(speed) * np.square(time) - np.square(distance)),
        ]
        return math.prod(_funcs)


@dataclass
class Weibull2Parameter(WeibullPDF):
    shape: float = 1.0
    scale: float = 1.0


@dataclass
class Weibull3Parameter(WeibullPDF):
    # Weibull parameterization using scale=a(t^b)(c^t)
    shape: float = 1.0
    a: float = 1.0
    b: float = 1.0
    c: float = 1.0


def calculate_etd_range(
    trajectory_gdf: Trajectory,
    output_path: typing.Union[str, bytes, os.PathLike],
    max_speed_kmhr: float = 0.0,
    max_speed_percentage: float = 0.9999,
    raster_profile: raster.RasterProfile = None,
    expansion_factor: float = 1.3,
    weibull_pdf: typing.Union[Weibull2Parameter, Weibull3Parameter] = Weibull2Parameter(),
) -> None:
    """
    The ETDRange class provides a trajectory-based, nonparametric approach to estimate the utilization distribution (UD)
    of an animal, using model parameters derived directly from the movement behaviour of the species.
    The model builds on the theory of "time-geography" whereby elliptical constrain- ing regions are established
    between temporally adjacent recorded locations.

    Parameters
    ----------
    trajectory_gdf : geopandas.GeoDataFrame
    output_path : str or PathLike
    max_speed_kmhr : float
    max_speed_percentage : 0.999
    raster_profile : raster.RasterProfile
    expansion_factor : float
    weibull_pdf : Weibull2Parameter or Weibull3Parameter

    Returns
    -------
    output_path : str
    """

    # if two-parameter weibull has default values; run an optimization routine to auto-determine parameters
    if isinstance(weibull_pdf, Weibull2Parameter) and all([weibull_pdf.shape == 1.0, weibull_pdf.scale == 1.0]):
        speed_kmhr = trajectory_gdf.speed_kmhr
        shape, scale = weibull_pdf.fit(speed_kmhr)

        # update the shape/scale parameters
        weibull_pdf.shape = shape
        weibull_pdf.scale = scale

    # reproject trajseg to desired crs
    trajectory_gdf = trajectory_gdf.to_crs(raster_profile.crs)

    # determine envelope of trajectory
    x_min, y_min, x_max, y_max = trajectory_gdf.geometry.total_bounds

    # apply expansion factor on the trajectory total bound.
    if expansion_factor > 1.0:
        dx = (x_max - x_min) * (expansion_factor - 1.0) / 2.0
        dy = (y_max - y_min) * (expansion_factor - 1.0) / 2.0
        x_min -= dx
        x_max += dx
        y_min -= dy
        y_max += dy

    # update the raster extent for the raster profile
    raster_profile.raster_extent = raster.RasterExtent(x_min=x_min, x_max=x_max, y_min=y_min, y_max=y_max)

    # determine the output raster size
    num_rows, num_columns = (
        raster_profile.rows,
        raster_profile.columns,
    )

    # maximum trajectory segment speed value
    max_trajseg_speed = trajectory_gdf.speed_kmhr.max()

    # determine max-speed value
    if max_speed_kmhr > 0.0:
        maxspeed = max_speed_kmhr
    else:
        # Use a value calculated from the CDF
        maxspeed = weibull_pdf.scale * math.pow(
            -1 * math.log(1.0 - max_speed_percentage),
            1.0 / weibull_pdf.shape,
        )
    if maxspeed <= max_trajseg_speed:
        raise ValueError(
            f"ETD maximum speed value: {maxspeed} should be greater than "
            f"trajectory maximum speed value {max_trajseg_speed}"
        )

    # Define the affine transform to get grid pixel centroids as geographic coordinates.
    grid_centroids = np.array(raster_profile.transform.to_gdal()).reshape(2, 3)
    grid_centroids[0, 0] = x_min + raster_profile.pixel_size * 0.5
    grid_centroids[1, 0] = y_max - raster_profile.pixel_size * 0.5

    centroids_coords = np.dot(grid_centroids, np.mgrid[1:2, :num_columns, :num_rows].T.reshape(-1, 3, 1))

    tr = sklearn.neighbors.KDTree(centroids_coords.squeeze().T)

    del centroids_coords

    points = np.stack(trajectory_gdf.geometry.apply(lambda x: np.array(x.coords.xy)))

    time = trajectory_gdf.timespan_seconds.values / 3600
    r = maxspeed * time * 1000

    start = tr.query_radius(points[:, :, 0], r=r, return_distance=True)
    end = tr.query_radius(points[:, :, 1], r=r, return_distance=True)

    del tr, points, r

    shape = weibull_pdf.shape
    scale = weibull_pdf.scale

    x = np.arange(0.001, maxspeed, 0.001)
    y = (4 * shape * scale ** (-shape) / np.pi) * np.array(
        [scipy.integrate.quad(_etd, m, maxspeed, args=(shape, scale, m))[0] for m in x]
    )

    raster_ndarray = np.zeros(num_rows * num_columns, dtype=np.float64)

    for k in range(len(start[0])):
        a, b, c = np.intersect1d(start[0][k], end[0][k], return_indices=True)
        speeds = (start[1][k][b] + end[1][k][c]) * 0.001 / time[k]
        i = speeds < maxspeed

        vals = y[np.digitize(speeds[i], x[:-1])] / time[k]

        raster_ndarray[a[i]] += vals / vals.sum() / time[k]

    # Normalize the grid values
    raster_ndarray = raster_ndarray / raster_ndarray.sum()

    # Set the null data values
    raster_ndarray[raster_ndarray == 0] = raster_profile.nodata_value

    # write raster_ndarray to GeoTIFF file.
    raster.RasterPy.write(
        ndarray=raster_ndarray.reshape(num_rows, num_columns),
        fp=output_path,
        **raster_profile,
    )



import geopandas as gpd
import geopandas.testing
import numpy as np
import pandas as pd
import pandas.testing
import pytest

import ecoscope


@pytest.mark.skipif(not pytest.earthranger, reason="No connection to EarthRanger")
def test_trajectory_is_not_empty(er_io):
    # test there is actually data in trajectory
    relocations = er_io.get_subjectgroup_observations(group_name=er_io.GROUP_NAME)
    trajectory = ecoscope.base.Trajectory.from_relocations(relocations)
    assert not trajectory.empty


@pytest.mark.skipif(not pytest.earthranger, reason="No connection to EarthRanger")
def test_redundant_columns_in_trajectory(er_io):
    # test there is no redundant column in trajectory
    relocations = er_io.get_subjectgroup_observations(group_name=er_io.GROUP_NAME)
    trajectory = ecoscope.base.Trajectory.from_relocations(relocations)
    assert "extra__fixtime" not in trajectory
    assert "extra___fixtime" not in trajectory
    assert "extra___geometry" not in trajectory


@pytest.mark.skipif(not pytest.earthranger, reason="No connection to EarthRanger")
def test_relocs_speedfilter(er_io):
    relocations = er_io.get_subjectgroup_observations(group_name=er_io.GROUP_NAME)
    relocs_speed_filter = ecoscope.base.RelocsSpeedFilter(max_speed_kmhr=8)
    relocs_after_filter = relocations.apply_reloc_filter(relocs_speed_filter)
    relocs_after_filter.remove_filtered(inplace=True)
    assert relocations.shape[0] != relocs_after_filter.shape[0]


@pytest.mark.skipif(not pytest.earthranger, reason="No connection to EarthRanger")
def test_relocs_distancefilter(er_io):
    relocations = er_io.get_subjectgroup_observations(group_name=er_io.GROUP_NAME)
    relocs_speed_filter = ecoscope.base.RelocsDistFilter(min_dist_km=1.0, max_dist_km=6.0)
    relocs_after_filter = relocations.apply_reloc_filter(relocs_speed_filter)
    relocs_after_filter.remove_filtered(inplace=True)
    assert relocations.shape[0] != relocs_after_filter.shape[0]


@pytest.mark.skipif(not pytest.earthranger, reason="No connection to EarthRanger")
def test_relocations_from_gdf_preserve_fields(er_io):
    relocations = er_io.get_subjectgroup_observations(group_name=er_io.GROUP_NAME)
    gpd.testing.assert_geodataframe_equal(relocations, ecoscope.base.Relocations.from_gdf(relocations))


def test_displacement_property(movbank_relocations):
    trajectory = ecoscope.base.Trajectory.from_relocations(movbank_relocations)
    expected = pd.Series(
        [2633.760505, 147749.545621],
        index=pd.Index(["Habiba", "Salif Keita"], name="groupby_col"),
    )
    pd.testing.assert_series_equal(
        trajectory.groupby("groupby_col").apply(ecoscope.base.Trajectory.get_displacement),
        expected,
    )


def test_tortuosity(movbank_relocations):
    trajectory = ecoscope.base.Trajectory.from_relocations(movbank_relocations)
    expected = pd.Series(
        [51.65388458528601, 75.96149479123005],
        index=pd.Index(["Habiba", "Salif Keita"], name="groupby_col"),
    )
    pd.testing.assert_series_equal(
        trajectory.groupby("groupby_col").apply(ecoscope.base.Trajectory.get_tortuosity),
        expected,
    )


def test_turn_angle(movbank_relocations):
    trajectory = ecoscope.base.Trajectory.from_relocations(movbank_relocations)
    trajectory = trajectory.loc[trajectory.groupby_col == "Habiba"].head(5)
    trajectory["heading"] = [0, 90, 120, 60, 300]
    turn_angle = trajectory.get_turn_angle()

    expected = pd.Series(
        [np.nan, 90, 30, -60, -120],
        dtype=np.float64,
        index=pd.Index([368706890, 368706891, 368706892, 368706893, 368706894], name="event-id"),
        name="turn_angle",
    )
    pandas.testing.assert_series_equal(turn_angle, expected)

    # Test filtering by dropping a row with index: 368706892.
    trajectory.drop(368706892, inplace=True)
    turn_angle = trajectory.get_turn_angle()
    expected = pd.Series(
        [np.nan, 90, np.nan, -120],
        dtype=np.float64,
        index=pd.Index([368706890, 368706891, 368706893, 368706894], name="event-id"),
        name="turn_angle",
    )

    pandas.testing.assert_series_equal(turn_angle, expected)


def test_sampling(movbank_relocations):
    relocs_1 = ecoscope.base.Relocations.from_gdf(
        gpd.GeoDataFrame(
            {"fixtime": pd.date_range(0, periods=1000, freq="1S", tz="utc")},
            geometry=gpd.points_from_xy(x=np.zeros(1000), y=np.linspace(0, 1, 1000)),
            crs=4326,
        )
    )
    traj_1 = ecoscope.base.Trajectory.from_relocations(relocs_1).loc[::2]
    upsampled_noncontiguous_1 = traj_1.upsample("3600S")

    relocs_2 = ecoscope.base.Relocations.from_gdf(
        gpd.GeoDataFrame(
            {"fixtime": pd.date_range(0, periods=10000, freq="2S", tz="utc")},
            geometry=gpd.points_from_xy(x=np.zeros(10000), y=np.linspace(0, 1, 10000)),
            crs=4326,
        )
    )
    traj_2 = ecoscope.base.Trajectory.from_relocations(relocs_2).loc[::2]
    upsampled_noncontiguous_2 = traj_2.upsample("3S")

    pnts_filter = ecoscope.base.RelocsCoordinateFilter(
        min_x=-5,
        max_x=1,
        min_y=12,
        max_y=18,
        filter_point_coords=[[180, 90], [0, 0]],
    )
    movbank_relocations.apply_reloc_filter(pnts_filter, inplace=True)
    movbank_relocations.remove_filtered(inplace=True)
    trajectory = ecoscope.base.Trajectory.from_relocations(movbank_relocations)
    downsampled_relocs_noint = trajectory.downsample("10800S", tolerance="900S")
    downsampled_relocs_int = trajectory.downsample("10800S", interpolation=True)

    expected_noncontiguous_1 = gpd.read_feather("tests/test_output/upsampled_noncontiguous_1.feather")
    expected_noncontiguous_2 = gpd.read_feather("tests/test_output/upsampled_noncontiguous_2.feather")
    expected_downsample_noint = gpd.read_feather("tests/test_output/downsampled_relocs_noint.feather")
    expected_downsample_int = gpd.read_feather("tests/test_output/downsampled_relocs.feather")

    gpd.testing.assert_geodataframe_equal(upsampled_noncontiguous_1, expected_noncontiguous_1, check_less_precise=True)
    gpd.testing.assert_geodataframe_equal(upsampled_noncontiguous_2, expected_noncontiguous_2, check_less_precise=True)
    gpd.testing.assert_geodataframe_equal(downsampled_relocs_noint, expected_downsample_noint, check_less_precise=True)
    gpd.testing.assert_geodataframe_equal(downsampled_relocs_int, expected_downsample_int, check_less_precise=True)


@pytest.mark.skip(reason="Intermittently fails. Need to fix in future PR")
@pytest.mark.filterwarnings("ignore:Target with index", 'ignore: ERFA function "utctai"')
def test_daynight_ratio(movbank_relocations):
    trajectory = ecoscope.base.Trajectory.from_relocations(movbank_relocations)
    expected = pd.Series(
        [
            2.212816,
            0.638830,
        ],
        index=pd.Index(["Habiba", "Salif Keita"], name="groupby_col"),
    )
    pd.testing.assert_series_equal(
        trajectory.groupby("groupby_col").apply(ecoscope.base.Trajectory.get_daynight_ratio),
        expected,
    )



import pandas as pd
import pytest

from ecoscope.analysis import immobility


def test_immobility(movbank_relocations):
    movbank_relocations = movbank_relocations.loc[movbank_relocations.groupby_col == "Salif Keita"][:100]
    immobility_profile = immobility.ImmobilityProfile(
        threshold_time=130, threshold_probability=0.5, threshold_radius=1000
    )
    result = immobility.Immobility.calculate_immobility(
        immobility_profile=immobility_profile, relocs=movbank_relocations
    )
    expected_result = {
        "probability_value": 0.83,
        "cluster_radius": 1913.5975702951566,
        "cluster_fix_count": 83,
        "total_fix_count": 100,
        "immobility_time": pd.Timedelta("4 days 03:00:00"),
    }

    assert result == pytest.approx(expected_result)



from math import ceil, floor

import geopandas as gpd
import igraph
import networkx as nx
import numpy as np
import pandas as pd
import rasterio
import sklearn.base
from affine import Affine
from shapely.geometry import shape
from skimage.draw import line

import ecoscope


class Ecograph:
    """
    A class that analyzes movement tracking data using Network Theory.

    Parameters
    ----------
    trajectory : ecoscope.base.Trajectory
        Trajectory dataframe
    resolution : float
        Pixel size, in meters
    radius : int
        Radius to compute Collective Influence (Default : 2)
    cutoff : int
        Cutoff to compute an approximation of betweenness index if the standard algorithm is
        too slow. Can be useful for very large graphs (Default : None)
    tortuosity_length : int
        The number of steps used to compute the two tortuosity metrics (Default : 3 steps)
    """

    def __init__(self, trajectory, resolution=15, radius=2, cutoff=None, tortuosity_length=3):
        self.graphs = {}
        self.trajectory = trajectory
        self.resolution = ceil(resolution)

        self.utm_crs = trajectory.estimate_utm_crs()
        self.trajectory.to_crs(self.utm_crs, inplace=True)
        self.features = [
            "dot_product",
            "speed",
            "step_length",
            "sin_time",
            "cos_time",
            "weight",
            "degree",
            "betweenness",
            "collective_influence",
            "tortuosity_1",
            "tortuosity_2",
        ]
        geom = self.trajectory["geometry"]

        eastings = np.array([geom.iloc[i].coords.xy[0] for i in range(len(geom))]).flatten()
        northings = np.array([geom.iloc[i].coords.xy[1] for i in range(len(geom))]).flatten()

        self.xmin = floor(np.min(eastings)) - self.resolution
        self.ymin = floor(np.min(northings)) - self.resolution
        self.xmax = ceil(np.max(eastings)) + self.resolution
        self.ymax = ceil(np.max(northings)) + self.resolution

        self.xmax += self.resolution - ((self.xmax - self.xmin) % self.resolution)
        self.ymax += self.resolution - ((self.ymax - self.ymin) % self.resolution)

        self.transform = Affine(self.resolution, 0.00, self.xmin, 0.00, -self.resolution, self.ymax)
        self.inverse_transform = ~self.transform

        self.n_rows = int((self.xmax - self.xmin) // self.resolution)
        self.n_cols = int((self.ymax - self.ymin) // self.resolution)

        def compute(df):
            subject_name = df.name
            print(f"Computing EcoGraph for subject {subject_name}")
            G = self._get_ecograph(self, df, subject_name, radius, cutoff, tortuosity_length)
            self.graphs[subject_name] = G

        self.trajectory.groupby("groupby_col").progress_apply(compute)

    def to_csv(self, output_path):
        """
        Saves the features of all nodes in a CSV file

        Parameters
        ----------
        output_path : str, Pathlike
            Output path for the CSV file
        """

        features_id = ["individual_name", "grid_id"] + self.features
        df = {feat_id: [] for feat_id in features_id}
        for individual_name, G in self.graphs.items():
            for node in G.nodes():
                df["individual_name"].append(individual_name)
                df["grid_id"].append(node)
                for feature in self.features:
                    df[feature].append(G.nodes[node][feature])
        (pd.DataFrame.from_dict(df)).to_csv(output_path, index=False)

    def to_geotiff(self, feature, output_path, individual="all", interpolation=None, transform=None):
        """
        Saves a specific node feature as a GeoTIFF

        Parameters
        ----------
        feature : str
            Feature of interest
        output_path : str, Pathlike
            Output path for the GeoTIFF file
        individual : str
            The individual for which we want to output the node feature (Default : "all")
        interpolation : str or None
            Whether to interpolate the feature for each step in the trajectory (Default : None). If
            provided, has to be one of those four types of interpolation : "mean", "median", "max"
            or "min"
        transform : sklearn.base.TransformerMixin or None
            A feature transform method (Default : None)
        """

        if feature in self.features:
            if individual == "all":
                feature_ndarray = self._get_feature_mosaic(self, feature, interpolation)
            elif individual in self.graphs.keys():
                feature_ndarray = self._get_feature_map(self, feature, individual, interpolation)
            else:
                raise IndividualNameError("This individual is not in the dataset")
        else:
            raise FeatureNameError("This feature was not computed by EcoGraph")

        if isinstance(transform, sklearn.base.TransformerMixin):
            nan_mask = ~np.isnan(feature_ndarray)
            feature_ndarray[nan_mask] = transform.fit_transform(feature_ndarray[nan_mask].reshape(-1, 1)).reshape(
                feature_ndarray[nan_mask].shape
            )

        raster_profile = ecoscope.io.raster.RasterProfile(
            pixel_size=self.resolution,
            crs=self.utm_crs,
            nodata_value=np.nan,
            band_count=1,
        )
        raster_profile.raster_extent = ecoscope.io.raster.RasterExtent(
            x_min=self.xmin, x_max=self.xmax, y_min=self.ymin, y_max=self.ymax
        )
        ecoscope.io.raster.RasterPy.write(
            ndarray=feature_ndarray,
            fp=output_path,
            **raster_profile,
        )

    @staticmethod
    def _get_ecograph(self, trajectory_gdf, individual_name, radius, cutoff, tortuosity_length):
        G = nx.Graph()
        geom = trajectory_gdf["geometry"]
        for i in range(len(geom) - (tortuosity_length - 1)):
            step_attributes = trajectory_gdf.iloc[i]
            lines = [list(geom.iloc[i + j].coords) for j in range(tortuosity_length)]
            p1, p2, p3, p4 = lines[0][0], lines[0][1], lines[1][1], lines[1][0]
            pixel1, pixel2 = (
                self.inverse_transform * p1,
                self.inverse_transform * p2,
            )

            row1, row2 = floor(pixel1[0]), floor(pixel2[0])
            col1, col2 = ceil(pixel1[1]), ceil(pixel2[1])

            t = step_attributes["segment_start"]
            seconds_in_day = 24 * 60 * 60
            seconds_past_midnight = (t.hour * 3600) + (t.minute * 60) + t.second + (t.microsecond / 1000000.0)
            time_diff = pd.to_datetime(
                trajectory_gdf.iloc[i + (tortuosity_length - 1)]["segment_end"]
            ) - pd.to_datetime(t)
            time_delta = time_diff.total_seconds() / 3600.0
            tortuosity_1, tortuosity_2 = self._get_tortuosities(self, lines, time_delta)

            attributes = {
                "dot_product": self._get_dot_product(p1, p2, p3, p4),
                "speed": step_attributes["speed_kmhr"],
                "step_length": step_attributes["dist_meters"],
                "sin_time": np.sin(2 * np.pi * seconds_past_midnight / seconds_in_day),
                "cos_time": np.cos(2 * np.pi * seconds_past_midnight / seconds_in_day),
                "tortuosity_1": tortuosity_1,
                "tortuosity_2": tortuosity_2,
            }

            if G.has_node((row1, col1)):
                self._update_node(G, (row1, col1), attributes)
            else:
                self._initialize_node(G, (row1, col1), attributes)
            if not G.has_node((row2, col2)):
                self._initialize_node(G, (row2, col2), attributes, empty=True)
            if (row1, col1) != (row2, col2):
                G.add_edge((row1, col1), (row2, col2))

        for node in G.nodes():
            for key in attributes.keys():
                if len(G.nodes[node][key]) != 0:
                    G.nodes[node][key] = np.nanmean(G.nodes[node][key])
                else:
                    G.nodes[node][key] = np.nan

        self._compute_network_metrics(self, G, radius, cutoff)
        return G

    @staticmethod
    def _update_node(G, node_id, attributes):
        G.add_node(node_id)
        G.nodes[node_id]["weight"] += 1
        for key, value in attributes.items():
            G.nodes[node_id][key].append(value)

    @staticmethod
    def _get_day_night_value(day_night_value):
        if day_night_value == "day":
            return 0
        elif day_night_value == "night":
            return 1

    @staticmethod
    def _initialize_node(G, node_id, attributes, empty=False):
        G.add_node(node_id)
        G.nodes[node_id]["weight"] = 1
        for key, value in attributes.items():
            if empty:
                G.nodes[node_id][key] = []
            else:
                G.nodes[node_id][key] = [value]

    @staticmethod
    def _get_dot_product(x, y, z, w):
        if (floor(y[0]) == floor(w[0])) and (floor(y[1]) == floor(w[1])):
            v = [y[0] - x[0], y[1] - x[1]]
            w = [z[0] - y[0], z[1] - y[1]]
            angle = np.arctan2(w[1], w[0]) - np.arctan2(v[1], v[0])

            while angle <= -np.pi:
                angle = angle + 2 * np.pi
            while angle > np.pi:
                angle = angle - 2 * np.pi
            return np.cos(angle)
        else:
            return np.nan

    @staticmethod
    def _get_tortuosities(self, lines, time_delta):
        point1, point2 = lines[0][0], lines[len(lines) - 1][1]
        beeline_dist = np.sqrt((point2[0] - point1[0]) ** 2 + (point2[1] - point1[1]) ** 2)
        total_length = 0
        for i in range(len(lines) - 1):
            point1, point2, point3 = lines[i][0], lines[i][1], lines[i + 1][0]
            if (floor(point2[0]) == floor(point3[0])) and (floor(point2[1]) == floor(point3[1])):
                total_length += ((point2[0] - point1[0]) ** 2 + (point2[1] - point1[1]) ** 2) ** 0.5
            else:
                if beeline_dist != 0:
                    return np.nan, np.log(time_delta / (beeline_dist**2))
                else:
                    return np.nan, np.nan
        point1, point2 = lines[len(lines) - 1][0], lines[len(lines) - 1][1]
        total_length += ((point2[0] - point1[0]) ** 2 + (point2[1] - point1[1]) ** 2) ** 0.5

        if (total_length != 0) and (beeline_dist != 0):
            return (beeline_dist / total_length), np.log(time_delta / (beeline_dist**2))
        elif (total_length == 0) and (beeline_dist > 0):
            return np.nan, np.log(time_delta / (beeline_dist**2))
        elif (total_length > 0) and (beeline_dist == 0):
            return 0, np.nan
        else:
            return np.nan, np.nan

    @staticmethod
    def _compute_network_metrics(self, G, radius, cutoff):
        self._compute_degree(G)
        self._compute_betweenness(G, cutoff)
        self._compute_collective_influence(self, G, radius)

    @staticmethod
    def _compute_degree(G):
        for node in G.nodes():
            G.nodes[node]["degree"] = G.degree[node]

    @staticmethod
    def _compute_collective_influence(self, G, radius):
        for node in G.nodes():
            G.nodes[node]["collective_influence"] = self._get_collective_influence(G, node, radius)

    @staticmethod
    def _get_collective_influence(G, start, radius):
        sub_g = nx.generators.ego_graph(G, start, radius=radius)
        collective_influence = 0
        for n in sub_g.nodes():
            collective_influence += G.degree[n] - 1
        collective_influence -= G.degree[start]
        return (G.degree[start]) * collective_influence

    @staticmethod
    def _compute_betweenness(G, cutoff):
        g = igraph.Graph.from_networkx(G)
        btw_idx = g.betweenness(cutoff=cutoff)
        for v in g.vs:
            node = v["_nx_name"]
            G.nodes[node]["betweenness"] = btw_idx[v.index]

    @staticmethod
    def _get_feature_mosaic(self, feature, interpolation=None):
        features = []
        for individual in self.graphs.keys():
            features.append(self._get_feature_map(self, feature, individual, interpolation))
        mosaic = np.full((self.n_cols, self.n_rows), np.nan)
        for i in range(self.n_cols):
            for j in range(self.n_rows):
                values = []
                for feature_map in features:
                    grid_val = feature_map[i][j]
                    if not np.isnan(grid_val):
                        values.append(grid_val)
                if len(values) >= 2:
                    mosaic[i][j] = np.mean(values)
                elif len(values) == 1:
                    mosaic[i][j] = values[0]
        return mosaic

    @staticmethod
    def _get_feature_map(self, feature, individual, interpolation):
        if interpolation is not None:
            return self._get_interpolated_feature_map(self, feature, individual, interpolation)
        else:
            return self._get_regular_feature_map(self, feature, individual)

    @staticmethod
    def _get_regular_feature_map(self, feature, individual):
        feature_ndarray = np.full((self.n_cols, self.n_rows), np.nan)
        for node in self.graphs[individual].nodes():
            feature_ndarray[node[1]][node[0]] = (self.graphs[individual]).nodes[node][feature]
        return feature_ndarray

    @staticmethod
    def _get_interpolated_feature_map(self, feature, individual, interpolation):
        feature_ndarray = self._get_regular_feature_map(self, feature, individual)
        individual_trajectory = self.trajectory[self.trajectory["groupby_col"] == individual]
        geom = individual_trajectory["geometry"]
        idxs_dict = {}
        for i in range(len(geom)):
            line1 = list(geom.iloc[i].coords)
            p1, p2 = line1[0], line1[1]
            pixel1, pixel2 = self.inverse_transform * p1, self.inverse_transform * p2
            row1, row2 = floor(pixel1[0]), floor(pixel2[0])
            col1, col2 = ceil(pixel1[1]), ceil(pixel2[1])

            rr, cc = line(col1, row1, col2, row2)
            for j in range(len(rr)):
                if np.isnan(feature_ndarray[rr[j], cc[j]]):
                    if (rr[j], cc[j]) in idxs_dict:
                        idxs_dict[(rr[j], cc[j])].append(feature_ndarray[rr[0], cc[0]])
                    else:
                        idxs_dict[(rr[j], cc[j])] = [feature_ndarray[rr[0], cc[0]]]
        for key, value in idxs_dict.items():
            if interpolation == "max":
                feature_ndarray[key[0], key[1]] = np.max(value)
            elif interpolation == "mean":
                feature_ndarray[key[0], key[1]] = np.mean(value)
            elif interpolation == "median":
                feature_ndarray[key[0], key[1]] = np.median(value)
            elif interpolation == "min":
                feature_ndarray[key[0], key[1]] = np.min(value)
            else:
                raise InterpolationError("Interpolation type not supported by EcoGraph")
        return feature_ndarray


class InterpolationError(Exception):
    pass


class IndividualNameError(Exception):
    pass


class FeatureNameError(Exception):
    pass


def get_feature_gdf(input_path):
    """
    Convert a GeoTIFF feature map into a GeoDataFrame

    Parameters
    ----------
    input_path : str, Pathlike
        Input path for the GeoTIFF file
    """
    shapes = []
    with rasterio.open(input_path) as src:
        crs = src.crs.to_wkt()
        data_array = src.read(1).astype(np.float32)
        data_array[data_array == src.nodata] = np.nan
        shapes.extend(rasterio.features.shapes(data_array, transform=src.transform))

    data = {"value": [], "geometry": []}
    for geom, value in shapes:
        if not np.isnan(value):
            data["geometry"].append(shape(geom))
            data["value"].append(value)

    df = pd.DataFrame.from_dict(data)
    return gpd.GeoDataFrame(df, geometry=df.geometry, crs=crs)
